/*
 * Copyright (c) Sun Microsystems, Inc. 1993-1997. All rights reserved.
 *
 * License is granted to copy, to use, and to make and to use derivative works
 * for research and evaluation purposes, provided that Sun Microsystems is
 * acknowledged in all documentation pertaining to any such copy or
 * derivative work. Sun Microsystems grants no other licenses expressed or
 * implied. The Sun Microsystems  trade name should not be used in any
 * advertising without its written permission.
 *
 * SUN MICROSYSTEMS MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS
 * SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided "as is"
 * without express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this software.
 */

/*
 * Copyright (c) 1993-1997 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and the Network Research Group at
 *	Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#)cbq.c  1.11     97/04/23 SMI"

#include <sys/types.h>
#include <sys/conf.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/stream.h>
#include <sys/strsun.h>
#include <sys/errno.h>
#include <sys/modctl.h>
#include <sys/sunddi.h>
#include <sys/stat.h>
#include <sys/ddi.h>
#include <sys/dlpi.h>
#include <sys/stropts.h>
#include <sys/strlog.h>
#include <sys/tihdr.h>
#include <sys/time.h>
#include <sys/tiuser.h>
#include <sys/kstat.h>
#include <sys/kmem.h>

#include <sys/socket.h>
#include <net/route.h>
#include <sys/sockio.h>
#include <netinet/in.h>
#include <inet/common.h>
#include <inet/mi.h>
#include <inet/ip.h>
#include <inet/ip_multi.h>
#include <inet/ip_ire.h>
#include <netinet/ip_mroute.h>
#include <sys/ethernet.h>

#include <netinet/rm_class.h>
#include <netinet/pktsched.h>
#include <netinet/cbq.h>

/*
 * Local Data structures.
 */

static void 	*cbq_g_head = NULL;
static cbq_t 	*cbq_g_ill_head = NULL;
kmutex_t	cbq_g_lock;

/*
 * Forward Declarations.
 */
static int		cbq_open(queue_t *, dev_t *, int, int, cred_t *);
static int		cbq_close(queue_t *);
static int		cbq_wput(queue_t *, mblk_t *);
static int		cbq_wsrv(queue_t *);
static int		cbq_rput(queue_t *, mblk_t *);
static int		cbq_rsrv(queue_t *);
static void		cbq_ioctl(queue_t *, mblk_t *);
static void		cbq_ioc(queue_t *, mblk_t *);
static void		cbq_proto(queue_t *, mblk_t *);
static void		cbq_pktsched(void *arg, int);
static void		cbq_deque(queue_t *);

static int		cbq_set_enable(struct cbq_interface *, int);
static void		cbq_satisfied(struct cbqstate *cbqp);
static int		cbq_getstats(struct cbq_getstats *gs);
static int		_cbq_getstats(struct cbq_getstats *gs,
				cbq_mod_state_t *cbqp);
static int		cbq_add_filter(struct cbq_add_filter *);
static int		cbq_delete_filter(struct cbq_delete_filter *);
static int		cbq_add_class(struct cbq_add_class *);
static int		cbq_delete_class(struct cbq_delete_class *);
static int		cbq_clear_hierarchy(struct cbq_interface *);

static int		cbq_clear_interface(cbq_mod_state_t *);
static void		cbq_dealloc_filter(cbq_mod_state_t *, k_filter_cb_t *);
static int		is_a_parent(cbq_mod_state_t *, struct rm_class *);
static k_filter_cb_t 	*cbq_filth_to_filtp(cbq_mod_state_t *, u_long, int *);
static u_long		cbq_alloc_class_handle(cbq_mod_state_t *, int *);
static void		cbq_free_class_handle(cbq_mod_state_t *, u_long);
static struct rm_class  *cbq_clh_to_clp(cbq_mod_state_t *, u_long, int *);

static int		cbq_ill_init(queue_t *, cbq_t *);
cbq_t			*cbq_ill_lookup_on_name(char *, int);
static int		cbq_ill_delete_entry(cbq_t *dcqp);
static mblk_t		*cbqgetpacket(struct cbqstate *cbqp);
static int		is_a_parent_class(cbq_mod_state_t *cbqp,
				struct rm_class *cl);

/*
 * Local Data structures.
 */
static struct module_info info = {
	0x4351, "cbq", 1, INFPSZ, CBQ_HI_WATER, CBQ_LO_WATER
};

static struct qinit rinit = {
	cbq_rput, cbq_rsrv, cbq_open, cbq_close, NULL, &info, NULL
};

static struct qinit winit = {
	cbq_wput, cbq_wsrv, cbq_open, cbq_close, NULL, &info, NULL
};

/*
 * Global data structures.
 */
struct streamtab cbqinfo = {
	&rinit, &winit, (struct qinit *)NULL, (struct qinit *)NULL
};

/*
 * Streams Module/Driver local functions.
 */

static int
cbq_pktsched_init(cbq_mod_state_t *cbqp)
{
	pkt_sched_cb_t	cb;

	cb.pktsched_cb = cbq_pktsched;
	cb.pktsched_cb_arg = (void *)cbqp;

	return (pktsched_init(cbqp->cbq_q, &cb, &cbqp->is_pktsched_inited,
		&cbqp->cbq_lock, &cbqp->cv));
}

static int
cbq_pktsched_destroy(cbq_mod_state_t *cbqp)
{
	return (pktsched_teardown(cbqp->cbq_q, &cbqp->is_pktsched_destroyed,
		&cbqp->cbq_lock, &cbqp->cv));
}

/*
 * static int
 * cbq_open(queue_t *, dev_t *, int, int, cred_t *) - This routine
 *	handles both Module and driver open.  If we are doing a Module
 *	open, we are being pushed on the Network Interface stack.  If we
 *	doing a driver open, then /dev/cbq is being opened to perform
 *	ioctl's on various network interfaces.
 *
 *	Returns:	ENODEV, if we opened too many /dev/cbqs.
 *			EBUSY, if we the device is already open.
 *			0, if we are successful.
 */

static int
cbq_open(queue_t *q, dev_t *devp, int flag, int sflag, cred_t *credp)
{
	int		i, instance, err = 0;
	int		priv = false;
	queue_t		*downstreamq, *tq;
	cbq_t		*cqp;
	cbq_mod_state_t	*cbqp;
	char		*cp;

	ENTRY("cbq_open");

	if (sflag == CLONEOPEN) {
		instance = getminor(*devp);
		if (instance >= N_MAX_CBQ_DEVS) {
			err = ENODEV;
			goto cbq_err_ret;
		} else if (q->q_ptr) {
			err = EBUSY;
			goto cbq_err_ret;
		}
	}

	qprocson(q);

	/*
	 * We are either opening as a device or a module.  In the device
	 * case, this is a CBQ client and we allocate a device structure.
	 * If it is a module open, then this is control stream access to the
	 * DLPI device.  We allocate a cbq_ill_t as the instance data in this
	 * case.
	 */
	mutex_enter(&cbq_g_lock);
	downstreamq = WR(q)->q_next;
	if (err = mi_open_comm(&cbq_g_head, sizeof (cbq_t), q, devp,
			flag, sflag, credp)) {
		mutex_exit(&cbq_g_lock);
		goto cbq_err_ret;
	}

	/*
	 * Are we being opened as a module?  If so, we are being pushed
	 * onto the interface ahead of IP.
	 */

	if (sflag == MODOPEN) {
		/* XXX - this should never happen */
		if (!downstreamq) {
			mi_close_comm(&cbq_g_head, q);
			mutex_exit(&cbq_g_lock);
			err = ENXIO;
			goto cbq_err_ret;
		}

		cqp = (cbq_t *)q->q_ptr;
		cqp->cbq_id = CBQ_MODULE;
		cbqp = &cqp->cbq.cbq_ms.cbq_state;

		/*
		 * Initialize the Module state structure.
		 */
		if (err = cbq_ill_init(WR(q), cqp)) {
			mi_close_comm(&cbq_g_head, q);
			mutex_exit(&cbq_g_lock);
			goto cbq_err_ret;
		}

		/*
		 * Initialize the cbq_state structure.
		 */
		mutex_init(&cqp->cbq.cbq_ms.cbq_state.cbq_lock,
			"cbq lock", MUTEX_DEFAULT, NULL);
		cv_init(&cqp->cbq.cbq_ms.cbq_state.cv,
			"cbq cv", CV_DEFAULT, NULL);

		cbqlock(q);
		cqp->cbq.cbq_ms.cbq_state.cbq_q = q;
#ifdef CBQ_DEBUG
		cbqp->cbqstarts = 0;
		cbqp->cbqfrees = 0;
#endif

		cqp->cbq.cbq_ms.cbq_state.cbq_freeit = NULL;
		cqp->cbq.cbq_ms.cbq_state.cbq_retn.free_func = cbqfree;
		cqp->cbq.cbq_ms.cbq_state.cbq_retn.free_arg =
			(char *)&cqp->cbq.cbq_ms.cbq_state;

		cqp->cbq.cbq_ms.cbq_state.cbq_timeout_id = 0;
		cqp->cbq.cbq_ms.cbq_state.q_full = 0;
		cqp->cbq.cbq_ms.cbq_state.qlen = 0;

		cbqp->cbq_class_tbl = (struct rm_class **)kmem_zalloc(
			sizeof (struct rm_class *) * CBQ_MAX_CLASSES, KM_SLEEP);
		if (!cbqp->cbq_class_tbl) {
			err = ENOMEM;
			mi_close_comm(&cbq_g_head, q);
			mutex_exit(&cbq_g_lock);
			goto cbq_err_ret;
		}

		cbqp->is_pktsched_inited = 0;
		cbqp->is_pktsched_destroyed = 0;
		cbqp->use_esballoc = 0;
		cbqp->mac_addr_sz = 0;
		cbqp->cbq_filter_debug = 0;

		for (i = 0; i < CBQ_FILTER_TBL_SZ; i++)
			cbqp->filters[i] = NULL;

		cbqp->threshold = CBQ_VIRTUAL_IF_THRESH;

		cbqp->cbq_stats_enabled = 0;
		cbqp->cbq_stats_timeout_id = 0;
		cbqunlock(q);
	} else {
		cqp = (cbq_t *)q->q_ptr;
		cqp->cbq_id = CBQ_DEVICE;

		cqp->cbq.cbq_ds.cbq_open = true;
		cqp->cbq.cbq_ds.cbq_q = q;

		/*
		 * XXX - need to teardown all classes on all interfaces.
		 * Assumption is that rsvpd above opens device and keeps it
		 * open until rsvpd dies or shutsdown.
		 */
	}
	mutex_exit(&cbq_g_lock);

cbq_err_ret:
	EXIT("cbq_open");
	return (err);
}

/*
 * static int
 * cbq_close(queue_t *) - This functions closes the cbq Module or driver.
 *
 *	Returns:	0, on success.
 */

static int
cbq_close(queue_t *q)
{
	cbq_t		*mcqp, *cqp;
	struct cbqstate	*cbqp;
	int		result;

	ENTRY("cbq_close");

	qprocsoff(q);

	mutex_enter(&cbq_g_lock);
	cqp = (cbq_t *)q->q_ptr;

	switch (cqp->cbq_id) {
	case CBQ_DEVICE:
		qprocsoff(q);
		for (mcqp = cbq_g_ill_head; mcqp;
				mcqp = (cbq_t *)mcqp->cbq.cbq_ms.cbq_ill_next) {
			cbqp = &mcqp->cbq.cbq_ms.cbq_state;

			cbqlock(cbqp->cbq_q);
			cbqp->cbq_enabled = DISABLE;
			cbq_clear_interface(cbqp);
			cbqunlock(cbqp->cbq_q);
		}
		break;

	case CBQ_MODULE:
		cbqp = &cqp->cbq.cbq_ms.cbq_state;
		cbqlock(q);

		cbqp->cbq_enabled = DISABLE;
		untimeout(cbqp->cbq_timeout_id);
		cbq_clear_interface(cbqp);

		if (cbqp->cbq_stats_enabled) {
			if (cbqp->cbq_stats_timeout_id)
				untimeout(cbqp->cbq_stats_timeout_id);
			cbqp->cbq_stats_enabled = 0;
		}

		/*
		 * Remove cqp from the ill list.
		 */
		result = cbq_ill_delete_entry(cqp);
		if (!result)
			cmn_err(CE_WARN, "cbq_close: ill not deleted\n");

		/*
		 * Where done unlock and destroy lock.
		 */
		cbqunlock(q);
		mutex_destroy(&cbqp->cbq_lock);
		break;
	}

	/*
	 * mi_close_comm frees the instance data structure.
	 */
	mi_close_comm(&cbq_g_head, q);
	q->q_ptr = (char *)NULL;
	mutex_exit(&cbq_g_lock);

	EXIT("cbq_close");
	return (0);
}

static int
cbq_is_enabled(cbq_mod_state_t *cbqp)
{
	return (cbqp->cbq_enabled);
}

/*
 * static int
 * cbq_wput(queue_t *, mblk_t *mp) - This function handles streams message
 *	processing for both the cbq module and driver.
 *
 *	If we are the driver, then only M_IOCTL, M_IOCDATA and M_FLUSH
 *	messages are processed.  If we traveling down the module path,
 *	then IP wants to classify a packet or just send it on depending
 *	whether CBQ is enabled or not. All control messages such M_FLUSH
 *	and et al are just sent on always.
 *
 *	Returns:	0, for success.
 */

static int
cbq_wput(queue_t *q, mblk_t *mp)
{
	cbq_t		*cqp = (cbq_t *)q->q_ptr;
	struct cbqstate *cbqp;

	/*
	 * The streams module does not process any IOCTL's.  The streams
	 * driver is the only processing machine for IOCTL's.
	 */
	switch (cqp->cbq_id) {
	case CBQ_MODULE:
		cbqp = &cqp->cbq.cbq_ms.cbq_state;
		if (!cbq_is_enabled(cbqp)) {
			putq(q, mp);
			return (0);
		}

		/*
		 * There are two cases for passing IP packets between IP and
		 * and the data link layer.  DL_UNITDATA_REQ and M_DATA.
		 * DL_UNITDATA if for non-fastpath capable interfaces, while
		 * M_DATA is fastpath capable.
		 */
		switch (DB_TYPE(mp)) {
		case M_PROTO:
		case M_PCPROTO:
			/* Look for DL_UNITDATA_REQ */
			cbqlock(q);
			cbq_proto(q, mp);
			cbqunlock(q);
			break;

		case M_DATA:
			/* Fastpath */
			cbqlock(q);
			cbqenque(q, mp, IS_FAST_PATH);
			cbqunlock(q);
			break;
		}
		break;

	case CBQ_DEVICE:
		switch (mp->b_datap->db_type) {
		case M_FLUSH:
			if (*mp->b_rptr & FLUSHW)
				flushq(q, FLUSHDATA);
			if (*mp->b_rptr & FLUSHR) {
				*mp->b_rptr &= ~FLUSHW;
				qreply(q, mp);
			} else
				freemsg(mp);
			break;

		case M_IOCTL:
			cbq_ioctl(q, mp);
			break;

		case M_IOCDATA:
			/* IOCTL continuation. */
			if (mi_copy_state(q, mp, (mblk_t **)NULL) == -1) {
				/*
				 * The copy operation failed.  mi_copy_state
				 * already cleaned up, so we can just return.
				 */
				return (0);
			}

			if (MI_COPY_DIRECTION(mp) == MI_COPY_IN)
				cbq_ioc(q, mp);
			else
				mi_copyout(q, mp);
			break;

		default:
			freemsg(mp);
			break;
		}
		break;

	default:
		cmn_err(CE_WARN, "cbq_wput: Illegal device or module\n");
		freemsg(mp);
		return (0);
	}

	return (0);
}

/*
 * static int
 * cbq_wsrv(queue_t *) - this function is only used by the cbq module
 *	to pass data from this module to the module downstream from us.
 *
 *	Returns:	0, on success.
 */

static int
cbq_wsrv(queue_t *q)
{
	mblk_t			*mp;

	while ((mp = getq(q)) != NULL) {
		if (canputnext(q))
			putnext(q, mp);
		else {
			putbq(q, mp);
			return (0);
		}
	}
	return (0);
}

/*
 * Static int
 * cbq_rput(queue_t *, mblk_t *) - This functions passes the data upstream
 *	from the cbq Module or driver.
 *
 *	Returns:	0, for success.
 */

static int
cbq_rput(queue_t *q, mblk_t *mp)
{
	putq(q, mp);
	return (0);
}

static int
cbq_rsrv(queue_t *q)
{
	mblk_t		*hp, *mp;
	struct iocblk	*iocp;
	cbq_t		*cqp;
	cbq_mod_state_t	*cbqp;
	int		dtype, just_free_it;

	cqp = (cbq_t *)WR(q)->q_ptr;

	while ((mp = getq(q)) != NULL) {
		dtype = DB_TYPE(mp);
		switch (dtype) {
		case M_IOCACK:
		case M_IOCNAK:
			if (cqp->cbq_id == CBQ_MODULE) {
				cbqp = &cqp->cbq.cbq_ms.cbq_state;
				cbqlock(WR(q));
				/*
				 * Does the driver implement the fastpath
				 * IOCTL.
				 */
				iocp = (struct iocblk *)mp->b_rptr;
				switch (iocp->ioc_cmd) {
				case DL_IOC_HDR_INFO:
					if (dtype == M_IOCACK) {
						hp = mp->b_cont->b_cont;
						cbqp->mac_addr_sz = MBLKL(hp);
					} else
						cbqp->mac_addr_sz = 0;
					break;
				case DL_IOC_PKTSCHED_INIT:
					if (dtype == M_IOCACK)
						cbqp->is_pktsched_inited = 1;
					else
						cbqp->is_pktsched_inited = -1;
					cv_broadcast(&cbqp->cv);
					just_free_it = 1;
					break;
				case DL_IOC_PKTSCHED_TEARDOWN:
					if (dtype == M_IOCACK)
						cbqp->is_pktsched_destroyed = 1;
					else
						cbqp->is_pktsched_destroyed =
							-1;
					cv_broadcast(&cbqp->cv);
					just_free_it = 1;
					break;
				}
				cbqunlock(WR(q));
			}
			break;
		default:
			break;
		}

		if (just_free_it) {
			freemsg(mp);
			just_free_it = 0;
		} else {
			if (canputnext(q))
				putnext(q, mp);
			else {
				putbq(q, mp);
				return (0);
			}
		}
	}

	return (0);
}

/*
 * static void
 * cbq_ioctl(queue_t *, mblk_t *) - Start the processing of CBQ device
 *	ioctls.  The actual processing is done in cbq_ioc.
 *
 *	Returns:	NONE.
 */

static void
cbq_ioctl(queue_t *q, mblk_t *mp)
{
	struct copyreq		*cq;
	cbq_t			*cqp = (cbq_t *)q->q_ptr;
	struct iocblk		*iocp = (struct iocblk *)mp->b_rptr;
	int			error = 0, copy_in_size = 0;
	caddr_t			uaddr = (caddr_t)NULL;

	ENTRY("cbq_ioctl");

	if (!mp->b_cont) {
		error = EINVAL;
		goto cbq_ioctl_done;
	}

	/*
	 * All of the ioctl's execpt for CBQ_DUMP_ILL_LIST require
	 * COPYIN to happen.  The actual ioctl is processed in cbq_ioc
	 * once the ioctl data is sent down to the CBQ device via the
	 * M_IOCDATA message.
	 */

	switch (iocp->ioc_cmd) {
	case CBQ_GETSTATS:
		copy_in_size = sizeof (struct cbq_getstats);
		break;

	case CBQ_ENABLE:
	case CBQ_DISABLE:
	case CBQ_ENABLE_STATS:
	case CBQ_DISABLE_STATS:
	case CBQ_CLEAR_HIERARCHY:
		copy_in_size = sizeof (struct cbq_interface);
		break;

	case CBQ_ADD_FILTER:
		copy_in_size = sizeof (struct cbq_add_filter);
		break;

	case CBQ_DEL_FILTER:
		copy_in_size = sizeof (struct cbq_delete_filter);
		break;

	case CBQ_ADD_CLASS:
		copy_in_size = sizeof (struct cbq_add_class);
		break;

	case CBQ_DEL_CLASS:
		copy_in_size = sizeof (struct cbq_delete_class);
		break;

	default:
		error = EINVAL;
		break;
	}

cbq_ioctl_done:
	if (error) {
		if (mp->b_cont) {
			freemsg(mp->b_cont);
			mp->b_cont = NULL;
		}
		mp->b_datap->db_type = M_IOCNAK;
		iocp->ioc_count = 0;
		iocp->ioc_error = error;
		qreply(q, mp);
	} else
		mi_copyin(q, mp, nilp(char), copy_in_size);

	EXIT("cbq_ioctl");
	return;
}

/*
 * static void
 * cbq_ioc(queue_t *, mblk_t *) - this functions performs ioctl's on the
 *	cbq device.  This function is called as result of data being
 *	copied in form user address space.  The stream copies the data in
 *	and then sends a M_IOCDATA message downstream.
 *
 *	Returns:	NONE
 */

static void
cbq_ioc(queue_t *q, mblk_t *mp)
{
	caddr_t				addr;
	struct cbq_add_filter		*afp;
	struct cbq_delete_filter	*dfp;
	struct cbq_add_class		*acp;
	struct cbq_delete_class		*dcp;
	mblk_t				*bp;
	struct iocblk			*iocp = (struct iocblk *)mp->b_rptr;
	int				error = 0;

	ENTRY("cbq_ioc");
	if (!(bp = mp->b_cont) || !(bp = bp->b_cont)) {
		error = EPROTO;
		goto cbq_ioc_done;
	}

	/* Get pointer to ioctl data. */
	addr = (char *)bp->b_rptr;
	switch (iocp->ioc_cmd) {
	case CBQ_GETSTATS:
		error = cbq_getstats((struct cbq_getstats *)addr);
		break;

	case CBQ_ENABLE_STATS:
	case CBQ_DISABLE_STATS:
		break;

	case CBQ_ENABLE:
		error = cbq_set_enable((struct cbq_interface *)addr, ENABLE);
		break;

	case CBQ_DISABLE:
		error = cbq_set_enable((struct cbq_interface *)addr, DISABLE);
		break;

	case CBQ_ADD_FILTER:
		afp = (struct cbq_add_filter *)addr;
		if ((error = cbq_add_filter(afp)) == 0)
			goto cbq_ioc_copyout;
		break;

	case CBQ_DEL_FILTER:
		dfp = (struct cbq_delete_filter *)addr;
		error = cbq_delete_filter(dfp);
		break;

	case CBQ_ADD_CLASS:
		acp = (struct cbq_add_class *)addr;

		if ((error = cbq_add_class(acp)) == 0)
			goto cbq_ioc_copyout;
		break;

	case CBQ_DEL_CLASS:
		dcp = (struct cbq_delete_class *)addr;
		error = cbq_delete_class(dcp);
		break;

	case CBQ_CLEAR_HIERARCHY:
		error = cbq_clear_hierarchy((struct cbq_interface *)addr);
		break;

	default:
		error = EINVAL;
		break;
	}

cbq_ioc_done:
	mi_copy_done(q, mp, error);
	EXIT("cbq_ioc(done)");
	return;

cbq_ioc_copyout:
	mi_copyout(q, mp);
	EXIT("cbq_ioc(copyout)");
	return;
}

static void
cbq_proto(queue_t *q, mblk_t *mp)
{
	union DL_primitives	*dlp;
	u_long			prim;
	mblk_t			*nmp;

	dlp = (union DL_primitives *)mp->b_rptr;
	prim = dlp->dl_primitive;

	if (prim == DL_UNITDATA_REQ) {
		if (((nmp = mp->b_cont) != NULL) && (DB_TYPE(nmp) == M_DATA))
			cbqenque(q, mp, IS_NOT_FAST_PATH);
	} else
		putq(q, mp);
}

/*
 * static int
 * cbq_set_enable(struct cbq_enable *ep) - this function processed the
 *	ioctl request to enable class based queueing.  It searches the list
 *	of interfaces for the specified interface and then enables CBQ on
 *	that interface.
 *
 *	Returns:	0, for no error.
 *			ENXIO, for specified inteface not found.
 */

static int
cbq_set_enable(struct cbq_interface *ep, int enable)
{
	int 		error = 0;
	cbq_t		*cqp;
	cbq_mod_state_t	*cbqp;
	char		*ifacename;
	int		ifacelen;

	ENTRY("cbq_set_enable");

	ifacename = ep->cbq_ifacename;
	ifacelen = ep->cbq_ifacelen;

	mutex_enter(&cbq_g_lock);
	if ((cqp = cbq_ill_lookup_on_name(ifacename, ifacelen))) {
		cbqlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
		cbqp = &cqp->cbq.cbq_ms.cbq_state;

		switch (enable) {
		case ENABLE:
			if (!cbqp->ifnp.root_ || !cbqp->ifnp.default_ ||
					!cbqp->ifnp.ctl_) {
				cbqp->cbq_enabled = DISABLE;
				if (!cbqp->ifnp.root_)
					cmn_err(CE_WARN,
						"No Root Class for %s\n",
						ifacename);
				if (!cbqp->ifnp.default_)
					cmn_err(CE_WARN,
						"No Default Class for %s\n",
						ifacename);
				if (!cbqp->ifnp.ctl_)
					cmn_err(CE_WARN,
						"No Control Class for %s\n",
						ifacename);
				error = ENXIO;
			} else if (!cbq_pktsched_init(cbqp)) {
				cbqp->use_esballoc++;
				cbqp->ifnp.maxqueued_ = 1;
			}
			cbqp->cbq_enabled = enable;
			break;

		case DISABLE:
			if (!cbqp->use_esballoc)
				cbq_pktsched_destroy(cbqp);
			cbqp->cbq_enabled = enable;
			break;
		}
		cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
		cbq_satisfied(cbqp);
	} else
		error = ENXIO;
	mutex_exit(&cbq_g_lock);

	EXIT("cbq_set_enable");
	return (error);
}

/*
 * void
 * cbq_satisfied(cbqp) - Check CBQ to see how link-sharing is performing.
 *
 *	Returns: NONE
 */

static void
cbq_satisfied(struct cbqstate *cbqp)
{
#ifdef USE_HRTIME
	hrtime_t	now;
#else
	struct timeval	now;
#endif
#ifdef notyet
	mutex_enter(&cbqp->cbq_lock);
	if (cbqp->cbq_enabled) {
		RM_GETTIME(now);
		rmc_tl_satisfied(&cbqp->ifnp, &now);
		cbqp->cbq_tl_timeout_id = timeout(cbq_satisfied,
			(caddr_t)cbqp, CBQ_LS_TIMEOUT);
	} else
		cbqp->cbq_tl_timeout_id = 0;
	mutex_exit(&cbqp->cbq_lock);
#endif
}

static int
cbq_getstats(struct cbq_getstats *gs)
{
	int		error = 0;
	cbq_t		*cqp;
	cbq_mod_state_t	*cbqp;
	char		*ifacename;
	int		ifacelen;

	ifacename = gs->iface.cbq_ifacename;
	ifacelen = gs->iface.cbq_ifacelen;

	mutex_enter(&cbq_g_lock);
	if ((cqp = cbq_ill_lookup_on_name(ifacename, ifacelen)) == NULL) {
		error = EINVAL;
		mutex_exit(&cbq_g_lock);
		return (error);
	}
	mutex_exit(&cbq_g_lock);

	cbqp = &cqp->cbq.cbq_ms.cbq_state;
	cbqlock(cbqp->cbq_q);
	error = _cbq_getstats(gs, cbqp);
	cbqunlock(cbqp->cbq_q);

	return (error);
}

static int
_cbq_getstats(struct cbq_getstats *gs, cbq_mod_state_t *cbqp)
{
	int		nclasses = gs->nclasses, i, error = 0;
	struct rm_class *cl;
	caddr_t		uaddr;

	/* Get the User Address Space pointer. */
	uaddr = (caddr_t)gs->stats;

	/*
	 * Copyout the root class.
	 */
	cl = cbqp->ifnp.root_;
	cl->stats_.depth = cl->depth_;
	if (copyout((caddr_t)&cl->stats_, uaddr, sizeof (class_stats_t))) {
		error = EFAULT;
		goto _cbq_getstats_exit;
	}

	/*
	 * Copyout the default class.
	 */
	cl = cbqp->ifnp.default_;
	cl->stats_.depth = cl->depth_;
	uaddr += sizeof (class_stats_t);
	if (copyout((caddr_t)&cl->stats_, uaddr, sizeof (class_stats_t))) {
		error = EFAULT;
		goto _cbq_getstats_exit;
	}

	/*
	 * Copyout the control class.
	 */
	cl = cbqp->ifnp.ctl_;
	cl->stats_.depth = cl->depth_;
	uaddr += sizeof (class_stats_t);
	if (copyout((caddr_t)&cl->stats_, uaddr, sizeof (class_stats_t))) {
		error = EFAULT;
		goto _cbq_getstats_exit;
	}
	nclasses -= 3;

	/*
	 * Get the rest of the classes.
	 */
	for (i = 0; (i < CBQ_MAX_CLASSES) && nclasses; i++) {
		if (cbqp->cbq_class_tbl[i] != NULL) {
			cl = cbqp->cbq_class_tbl[i];
			cl->stats_.depth = cl->depth_;
			uaddr += sizeof (class_stats_t);
			if (copyout((caddr_t)&cl->stats_, uaddr,
					sizeof (class_stats_t))) {
				error = EFAULT;
				goto _cbq_getstats_exit;
			}
			nclasses--;
		}
	}

_cbq_getstats_exit:
	return (error);
}

static int
cbq_check_filter_handle(k_filter_cb_t *flist, u_int handle)
{
	k_filter_cb_t	*kfp;

	for (kfp = flist; kfp != NULL; kfp = kfp->next) {
		if (kfp->handle == handle)
			return (1);
	}

	return (0);
}

static int
cbq_add_filter(struct cbq_add_filter *afp)
{
	int			error = 0;
	u_int			fhi;
	char			*ifacename;
	int			i, ifacelen;
	cbq_t			*cqp;
	cbq_mod_state_t		*cbqp;
	k_filter_cb_t		*kfcbp, *kfp, *cl_filt_p;
	struct rm_class		*classp;

	u_int			hindex;
	hrtime_t		hrtime;

	ENTRY("cbq_add_filter");

	ifacename = afp->cbq_iface.cbq_ifacename;
	ifacelen = afp->cbq_iface.cbq_ifacelen;

	if ((cqp = cbq_ill_lookup_on_name(ifacename, ifacelen)) != NULL) {
		cbqlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
		cbqp = &cqp->cbq.cbq_ms.cbq_state;

		if (afp->cbq_filter.daddr == 0) {
			error = EINVAL;
			cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
			goto cbq_add_filter_exit;
		}

		/* Get the pointer to class. */
		classp = cbq_clh_to_clp(cbqp, afp->cbq_class_handle, &error);
		if (error) {
			cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
			goto cbq_add_filter_exit;
		}

		if ((kfcbp = kmem_zalloc(sizeof (k_filter_cb_t),
				KM_SLEEP)) == NULL) {
			error = ENOMEM;
			cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
			goto cbq_add_filter_exit;
		}

		kfcbp->filter = afp->cbq_filter;

		/*
		 * Generate Handle for filter.
		 */
		hindex = CBQ_GET_HASH_INDEX(kfcbp->filter.daddr);
cbq_get_handle:
		hrtime = gethrtime();
		kfcbp->handle = (hindex << 24) |
			(u_int)((hrtime & 0x00ffffff));
		if (cbq_check_filter_handle(cbqp->filters[hindex],
				kfcbp->handle))
			goto cbq_get_handle;

		kfcbp->clp = classp;

		/*
		 * Add filter to front of filter list.
		 */
		kfcbp->prev = NULL;
		kfcbp->next = cbqp->filters[hindex];
		if (cbqp->filters[hindex])
			cbqp->filters[hindex]->prev = kfcbp;
		cbqp->filters[hindex] = kfcbp;
		afp->cbq_filter_handle = kfcbp->handle;
		cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
	} else
		error = ENXIO;

cbq_add_filter_exit:
	EXIT("cbq_add_filter");
	return (error);
}

static int
cbq_delete_filter(struct cbq_delete_filter *dfp)
{
	int			error = 0;
	char			*ifacename;
	int			ifacelen;
	struct rm_class		*classp;
	cbq_t			*cqp;
	u_long			f_handle;
	cbq_mod_state_t		*cbqp;
	k_filter_cb_t		*kfp;

	ENTRY("cbq_delete_filter");

	ifacename = dfp->cbq_iface.cbq_ifacename;
	ifacelen = dfp->cbq_iface.cbq_ifacelen;

	if ((cqp = cbq_ill_lookup_on_name(ifacename, ifacelen)) != NULL) {
		cbqlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);

		cbqp = &cqp->cbq.cbq_ms.cbq_state;
		kfp = cbq_filth_to_filtp(cbqp, dfp->cbq_filter_handle, &error);
		if (error) {
			cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
			goto cbq_delete_filter_exit;
		}

		cbq_dealloc_filter(cbqp, kfp);
		cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
	} else
		error = ENXIO;

cbq_delete_filter_exit:
	EXIT("cbq_delete_filter");
	return (error);
}

static int
cbq_add_class(struct cbq_add_class *acp)
{
	int			error = 0;
	char			*ifacename;
	int			ifacelen;
	struct rm_class		*classp, *borrow, *parent;
	cbq_t			*cqp;
	cbq_mod_state_t		*cbqp;
	u_long			chandle;

	ENTRY("cbq_add_class");

	ifacename = acp->cbq_iface.cbq_ifacename;
	ifacelen = acp->cbq_iface.cbq_ifacelen;

	if ((cqp = cbq_ill_lookup_on_name(ifacename, ifacelen)) != NULL) {
		cbqlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
		cbqp = &cqp->cbq.cbq_ms.cbq_state;

		/* Get pointers to parent and borrow classes.  */
		parent = cbq_clh_to_clp(cbqp,
			acp->cbq_class.parent_class_handle, &error);
		if (error)  {
			cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
			goto cbq_add_class_exit;
		}

		borrow = cbq_clh_to_clp(cbqp,
			acp->cbq_class.borrow_class_handle, &error);
		if (error) {
			cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
			goto cbq_add_class_exit;
		}

		/*
		 * A class must borrow from it's parent.
		 */
		if (borrow != parent) {
			cmn_err(CE_CONT,
				"cbq_add_class: borrow class != parent\n");
			error = EINVAL;
			cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
			goto cbq_add_class_exit;

		}

		/* Create class.  */
		if (!parent) {
			if (cbqp->ifnp.root_) {
				error = EINVAL;
				cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
				goto cbq_add_class_exit;
			}

			chandle = ROOT_CLASS_HANDLE;
		} else if (acp->cbq_class.defaultclass) {
			if (cbqp->ifnp.default_) {
				error = EINVAL;
				cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
				goto cbq_add_class_exit;
			}

			chandle = DEFAULT_CLASS_HANDLE;
		} else if (acp->cbq_class.ctlclass) {
			if (cbqp->ifnp.ctl_) {
				error = EINVAL;
				cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
				goto cbq_add_class_exit;
			}
			chandle = CTL_CLASS_HANDLE;
		} else {
			chandle = cbq_alloc_class_handle(cbqp, &error);
			if (error) {
				cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
				goto cbq_add_class_exit;
			}
		}

		if ((classp = cbq_class_create(cbqp, acp, chandle,
				parent, borrow)) == NULL) {
			cbq_free_class_handle(cbqp, chandle);
			error = ENOMEM;
			cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
			goto cbq_add_class_exit;
		}

		/* Return handle to user space. */
		classp->stats_.handle = chandle;
		classp->stats_.depth = classp->depth_;
		acp->cbq_class_handle = chandle;
		cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
	} else
		error = ENXIO;

cbq_add_class_exit:
	EXIT("cbq_add_class");
	return (error);
}

static int
cbq_delete_class(struct cbq_delete_class *dcp)
{
	int			i, error = 0;
	char			*ifacename;
	int			ifacelen;
	struct rm_class		*classp;
	cbq_t			*cqp;
	cbq_mod_state_t		*cbqp;
	k_filter_cb_t		*kfp, *tfp;

	ENTRY("cbq_delete_class");

	ifacename = dcp->cbq_iface.cbq_ifacename;
	ifacelen = dcp->cbq_iface.cbq_ifacelen;

	if ((cqp = cbq_ill_lookup_on_name(ifacename, ifacelen)) != NULL) {
		cbqlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
		cbqp = &cqp->cbq.cbq_ms.cbq_state;

		classp = cbq_clh_to_clp(cbqp, dcp->cbq_class_handle, &error);
		if (error || !classp) {
			cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
			goto cbq_del_class_exit;
		}

		/*
		 * If a filter has a reference to this class delete
		 * the filter before deleting the class.
		 */
		for (i = 0; i < CBQ_FILTER_TBL_SZ; i++) {
			for (kfp = cbqp->filters[i]; kfp; kfp = tfp) {
				tfp = kfp->next;
				if (kfp->clp == classp)
					cbq_dealloc_filter(cbqp, kfp);
			}
		}

		/*
		 * If we are a parent class, then return an error.
		 */
		if (is_a_parent_class(cbqp, classp)) {
			error = EINVAL;
			cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
			goto cbq_del_class_exit;
		}

		error = cbq_class_destroy(cbqp, classp);
		if (error) {
			cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
			goto cbq_del_class_exit;
		}

		cbq_free_class_handle(cbqp, dcp->cbq_class_handle);
		cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);
	} else
		error = ENXIO;

cbq_del_class_exit:
	EXIT("cbq_delete_class");
	return (error);
}

static int
cbq_clear_hierarchy(struct cbq_interface *ifp)
{
	int		ifacelen, error = 0;
	char		*ifacename;
	cbq_t		*cqp;
	cbq_mod_state_t	*cbqp;

	ENTRY("cbq_clear_hierarchy");

	ifacename = ifp->cbq_ifacename;
	ifacelen = ifp->cbq_ifacelen;

	if ((cqp = cbq_ill_lookup_on_name(ifacename, ifacelen)) != NULL) {
		cbqlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);

		cbqp = &cqp->cbq.cbq_ms.cbq_state;
		cbqp->cbq_enabled = DISABLE;
		error = cbq_clear_interface(cbqp);
		cbqunlock(cqp->cbq.cbq_ms.cbq_state.cbq_q);

	} else
		error = ENXIO;

cbq_clear_hierarchy_exit:
	EXIT("cbq_clear_hierarchy");
	return (error);
}

static int
cbq_clear_interface(cbq_mod_state_t *cbqp)
{
	int		again, i;
	k_filter_cb_t	*kfp, *tfp;
	struct rm_class	*cl;

	/*
	 * Free the filters for this interface.
	 */
	for (i = 0; i < CBQ_FILTER_TBL_SZ; i++) {
		for (kfp = cbqp->filters[i]; kfp; kfp = tfp) {
			tfp = kfp->next;
			cbq_dealloc_filter(cbqp, kfp);
		}
	}

	/*
	 * Clear out the classes now.
	 */
do_it_again:
	again = 0;
	for (i = 0; i < CBQ_MAX_CLASSES; i++) {
		if ((cl = cbqp->cbq_class_tbl[i]) == NULL)
			continue;
		if (is_a_parent_class(cbqp, cl)) {
			again++;
			continue;
		}

		cbq_class_destroy(cbqp, cl);
		cbqp->cbq_class_tbl[i] = NULL;
	}
	if (again)
		goto do_it_again;

	if (cbqp->ifnp.ctl_) {
		cbq_class_destroy(cbqp, cbqp->ifnp.ctl_);
		cbqp->ifnp.ctl_ = NULL;
	}

	if (cbqp->ifnp.default_) {
		cbq_class_destroy(cbqp, cbqp->ifnp.default_);
		cbqp->ifnp.default_ = NULL;
	}

	if (cbqp->ifnp.root_) {
		cbq_class_destroy(cbqp, cbqp->ifnp.root_);
		cbqp->ifnp.root_ = NULL;
	}
	return (0);
}

static int
is_a_parent_class(cbq_mod_state_t *cbqp, struct rm_class *cl)
{
	int	i;

	if (!cl)
		return (0);

	for (i = 0; i < CBQ_MAX_CLASSES; i++) {
		if (cbqp->cbq_class_tbl[i]) {
			if (cbqp->cbq_class_tbl[i]->parent_ == cl)
				return (1);
		}
	}

	return (0);
}

static void
cbq_dealloc_filter(cbq_mod_state_t *cbqp, k_filter_cb_t *kfp)
{
	u_int	hindex;

	if (!kfp || !cbqp)
		return;

	/* Delete from the filter list. */
	hindex = CBQ_GET_HINDEX(kfp->handle);

	if (kfp->prev && kfp->next) {
		kfp->prev->next = kfp->next;
		kfp->next->prev = kfp->prev;
	} else if (kfp->prev && !kfp->next) {
		kfp->prev->next = NULL;
	} else if (kfp->next && !kfp->prev) {
		kfp->next->prev = NULL;
		cbqp->filters[hindex] = kfp->next;
	} else
		cbqp->filters[hindex] = NULL;

	kfp->next = kfp->prev = NULL;
	kmem_free((void *)kfp, sizeof (k_filter_cb_t));
}

static k_filter_cb_t *
cbq_filth_to_filtp(cbq_mod_state_t *cbqp, u_long handle, int *error)
{
	k_filter_cb_t 	*kfp;
	u_int		hindex;

	hindex = CBQ_GET_HINDEX(handle);
	if (cbqp->filters[hindex]) {
		for (kfp = cbqp->filters[hindex]; kfp; kfp = kfp->next) {
			if (kfp->handle == handle)
				return (kfp);
		}
	}

	*error = EINVAL;
	return ((k_filter_cb_t *)NULL);
}

static u_long
cbq_alloc_class_handle(cbq_mod_state_t *cbqp, int *error)
{
	int	i;
	u_long	handle = 0;

	for (i = 0; i < CBQ_MAX_CLASSES; i++) {
		if (!cbqp->cbq_class_tbl[i])
			break;
	}

	if (i >= CBQ_MAX_CLASSES)
		*error = ENOSPC;

	handle = (u_long)i;
	return (handle);
}

static void
cbq_free_class_handle(cbq_mod_state_t *cbqp, u_long handle)
{
	switch (handle) {
	case ROOT_CLASS_HANDLE:
	case DEFAULT_CLASS_HANDLE:
	case CTL_CLASS_HANDLE:
		break;

	default:
		if (handle >= CBQ_MAX_CLASSES)
			return;
		cbqp->cbq_class_tbl[handle] = NULL;
	}
}

static struct rm_class *
cbq_clh_to_clp(cbq_mod_state_t *cbqp, u_long handle, int *error)
{
	switch (handle) {
	case NULL_CLASS_HANDLE:
		return ((struct rm_class *)NULL);

	case ROOT_CLASS_HANDLE:
		if (!cbqp->ifnp.root_)
			*error = EINVAL;
		return (cbqp->ifnp.root_);

	case DEFAULT_CLASS_HANDLE:
		if (!cbqp->ifnp.default_)
			*error = EINVAL;
		return (cbqp->ifnp.default_);

	case CTL_CLASS_HANDLE:
		if (!cbqp->ifnp.ctl_)
			*error = EINVAL;
		return (cbqp->ifnp.ctl_);

	default:
		if (handle >= CBQ_MAX_CLASSES)
			*error = ENXIO;
		else
			return (cbqp->cbq_class_tbl[handle]);
	}
}

/*
 * static int
 * cbq_ill_init(queue_t *, cbq_t *) - this functions is called when CBQ is
 *	pushed onto the stream to the network interface.  This routine mainly
 *	gets the name of the network interface below it and adds it to a list
 *	network interfaces that CBQ has pushed onto.
 *
 *	Returns:	ENXIO, if no such interface exists.
 *			0, if no error occurred.
 */

static int
cbq_ill_init(queue_t *q, cbq_t *cqp)
{
	queue_t	*tq;
	char	*cp;
	int	count;

	ENTRY("cbq_ill_init");

	/*
	 * Don't look! We walk down the stream to get the name of the
	 * driver so we can constrcut a the ILL name for this module.
	 */
	for (tq = q; (tq && tq->q_next); tq = tq->q_next)
		;
	cp = tq->q_qinfo->qi_minfo->mi_idname;
	if (!cp || !*cp) {
		EXIT("cbq_ill_init");
		return (ENXIO);
	}

	/*
	 * Contruct a name from our devices 'q_qinfo->qi_minfo->mi_idname'
	 * and the lowest 'unit number' that is unused.  We assume here
	 * that devices are opened in ascending order.
	 */
	count = -1;
	do {
		if (++count > 1000) {
			EXIT("cbq_ill_init");
			return (ENXIO);
		}
		mi_sprintf(cqp->cbq.cbq_ms.cbq_ill_name, "%s%d", cp, count);
		cqp->cbq.cbq_ms.cbq_ill_name_length =
			mi_strlen(cqp->cbq.cbq_ms.cbq_ill_name);
	} while (cbq_ill_lookup_on_name(cqp->cbq.cbq_ms.cbq_ill_name,
			cqp->cbq.cbq_ms.cbq_ill_name_length));

#ifdef CBQ_DEBUG
	cmn_err(CE_CONT, "cbq_ill_init: interface %s\n",
		cqp->cbq.cbq_ms.cbq_ill_name);
#endif
	cqp->cbq.cbq_ms.cbq_ill_ppa = count;
	cqp->cbq.cbq_ms.cbq_ill_next = NULL;

	/*
	 * Append to the list of cbq_ill's.
	 */
	if (cbq_g_ill_head) {
		cbq_t	*illp;

		illp = cbq_g_ill_head;
		while (illp->cbq.cbq_ms.cbq_ill_next)
			illp = (cbq_t *)illp->cbq.cbq_ms.cbq_ill_next;

		illp->cbq.cbq_ms.cbq_ill_next = (struct cbq_ill_s *)cqp;
	} else
		cbq_g_ill_head = cqp;

	EXIT("cbq_ill_init");
	return (0);
}

/*
 * cbq_t *
 * cbq_ill_lookup_on_name(char *, int) - return the pointer to the cbq_t
 *	structure given given by name and namelen.
 *
 *	Returns:	NULL, if unable to find the interface.
 *			pointer to cbq_t structure.
 */

cbq_t *
cbq_ill_lookup_on_name(char *name, int namelen)
{
	cbq_t	*cqp = NULL;
	int	i;

	if (namelen == 0)
		return (cqp);

	for (cqp = cbq_g_ill_head; cqp;
			cqp = (cbq_t *)cqp->cbq.cbq_ms.cbq_ill_next) {
		if (!mi_strcmp(cqp->cbq.cbq_ms.cbq_ill_name, name))
			break;
	}

	return (cqp);
}

/*
 * static int
 * cbq_ill_delete_entry(cbq_t *dcqp)
 *
 * Deletes 'dcpq' from cbq_g_ill_head list, relinking correctly the list
 * itself.
 *
 * Returns:	0: Element not found.
 *		1: Element deleted.
 */

static int
cbq_ill_delete_entry(cbq_t *dcqp)
{
	cbq_t	*pcqp = NULL, *cqp = NULL;
	int	i;

	/*
	 * Find the entry in the list of 'ills'.
	 */
	for (cqp = cbq_g_ill_head; cqp;
			cqp = (cbq_t *)cqp->cbq.cbq_ms.cbq_ill_next) {
		if (cqp == dcqp)
			break;
		else
			pcqp = cqp;
	}

	if (!cqp)
		return (0);

	/*
	 * Update its neighbors.
	 */
	if (!pcqp)
		cbq_g_ill_head = (cbq_t *)dcqp->cbq.cbq_ms.cbq_ill_next;
	else
		pcqp->cbq.cbq_ms.cbq_ill_next = dcqp->cbq.cbq_ms.cbq_ill_next;
	dcqp->cbq.cbq_ms.cbq_ill_next = NULL;

	/* Return OK condition.  */
	return (1);
}

static mblk_t *
cbqgetpacket(struct cbqstate *cbqp)
{
	struct timeval		now;
	mblk_t			*mp;

	if (cbqp == NULL)
		return (NULL);

	if ((mp = rmc_dequeue_next(&cbqp->ifnp)) != NULL)
		return (mp);

	return (NULL);
}

/*
 * void
 * cbqenque(queue_t *q, mblk_t *mp, int fastpath) - Queue data packets.
 *
 *	Assumptions:	State lock help by this thread when entering.
 *	Returns:	NONE
 */

void
cbqenque(queue_t *q, mblk_t *mp, int fastpath)
{
	cbq_t		*cqp = (cbq_t *)q->q_ptr;
	struct cbqstate	*cbqp;
	mblk_t		*bp;

	/*
	 * If this message appears all bets are off
	 */
	if (mp->b_next) {
		cmn_err(CE_WARN, "cbqenque: n_next != NULL\n");
		mp->b_next = 0;
	}

	cbq_class_queue(q, mp, fastpath);

	/*
	 * XXX - This is not quite correct we shouldn't need to
	 * bootstrap the first sending of a packet like this
	 */
	cbq_deque(q);
}

/*
 * void
 * cbqstart(queue_t *q, mblk_t *mp, *struct cbqstate *cbqp)
 *
 *	As we are queueing messages ourselves and only passing messages down
 *	to the mac driver one message at a time we need to know when the
 *	previous packet has been transmitted.  We the the mac completion
 *	interrupt by using esballoc to allocate a new header.  Esballoc
 *	allows the caller to supply its own free routine.  So, whenever the
 *	free routine is called we know the packet has been put out on the
 *	wire.
 *
 *	Returns:	NONE
 */

void
cbqstart(queue_t *q, mblk_t *mp, struct cbqstate *cbqp)
{
	mblk_t		*bp, *nmp;
	u_long		packet_time;
	struct timeval	now;

	if (!cbqp)
		cmn_err(CE_PANIC, "cbqstart, state is NULL\n");

	/*
	 * If the device below does not implement pktsched hooks
	 * resort to using esballoc.
	 */
	if (cbqp->use_esballoc) {
		if ((nmp = esballoc(mp->b_rptr, mp->b_wptr - mp->b_rptr,
				BPRI_HI, &cbqp->cbq_retn)) == 0) {
			cmn_err(CE_WARN, "cbqstart: esballoc failed\n");
			freemsg(mp);
			return;
		}
#ifdef DEBUG
		cbqp->cbqstarts++;
#endif
		cbqp->cbq_freeit = mp;
		nmp->b_datap->db_type = mp->b_datap->db_type;
		nmp->b_cont = mp->b_cont;
		mp->b_cont = 0;
		nmp->b_wptr = mp->b_wptr;
		bp = nmp;
	} else
		bp = mp;

	putq(q, bp);
	qenable(q);
}

/*
 * void
 * cbq_pktsched(void *arg, int nbytes)
 *	This function is called when a device driver has put a packet
 *	onto the wire.
 */
void
cbq_pktsched(void *arg, int nbytes)
{
	cbqfree(arg);
}

/*
 * void
 * cbqfree(struct cbqstate *)
 *
 *	Free the message header and start the next transfer.  This is
 *	called upon completion of the data transfer of the packet.
 *
 *	Returns:	NONE
 */

void
cbqfree(void *arg)
{
	mblk_t		*mp;
	cbq_mod_state_t	*cbqp = (cbq_mod_state_t *)arg;
	queue_t		*q = cbqp->cbq_q;

	cbqlock(q);

	if (!cbqp->cbq_enabled) {
		cbqunlock(cbqp->cbq_q);
		return;
	}

	/*
	 * If we are using esballoc, then free the message.
	 */
	if (cbqp->use_esballoc && cbqp->cbq_freeit) {
		freemsg(cbqp->cbq_freeit);
		cbqp->cbq_freeit = NULL;
	}

#ifdef DEBUG
	cbqp->cbqfrees++;
#endif
	/* Update the class and schedule the next packet. */
	rmc_update_class_util(&cbqp->ifnp);

	/* Start the next packet, if there is one.  */
	cbq_deque(q);

	cbqunlock(q);
}

/*
 * void
 * cbq_deque(queue_t *) - Started sending of data.
 *	Returns:	NONE
 */

static void
cbq_deque(queue_t *q)
{
	mblk_t		*mp;
	cbq_t		*cqp = (cbq_t *)q->q_ptr;
	struct cbqstate	*cbqp = &cqp->cbq.cbq_ms.cbq_state;

	/*
	 * If we are inactive, then send out data if there is
	 * some to send.
	 */
	if ((mp = cbqgetpacket(cbqp)) != NULL) {
		mp->b_next = 0;
		cbqstart(WR(q), mp, cbqp);
	}
}

/*
 * void
 * cbqwatchdog(queue_t *q) - This function handles overlimit rm_class code
 *	callout.
 *
 *	Returns:	NONE
 */

void
cbqwatchdog(queue_t *q)
{
	cbq_t		*cqp = (cbq_t *)q->q_ptr;
	struct cbqstate	*cbqp = &cqp->cbq.cbq_ms.cbq_state;

	/*
	 * If the CBQ is enabled on this interface, make sure
	 * the it is running if there are packets to send.
	 */
	if (cbqp->cbq_enabled)
		cbq_deque(q);
}

/*
 * void
 * cbqlock(queue_t *) - grab lock for this cbq_t struture.
 *
 *	Retunrns:	NONE
 */

void
cbqlock(queue_t *q)
{
	cbq_t	*cqp = (cbq_t *)q->q_ptr;

	if (cqp && (cqp->cbq_id == CBQ_MODULE))
		mutex_enter(&cqp->cbq.cbq_ms.cbq_state.cbq_lock);
	else if (cqp && (cqp->cbq_id == CBQ_DEVICE)) {
		cmn_err(CE_PANIC,
			"cbqlock: cbq device attempting MODULE operation\n");
	}
}

/*
 * void
 * cbqunlock(queue_t *) - let lock go
 *
 *	Returns:	NONE
 */

void
cbqunlock(queue_t *q)
{
	cbq_t	*cqp = (cbq_t *)q->q_ptr;

	if (cqp && (cqp->cbq_id == CBQ_MODULE))
		mutex_exit(&cqp->cbq.cbq_ms.cbq_state.cbq_lock);
	else if (cqp && (cqp->cbq_id == CBQ_DEVICE)) {
		cmn_err(CE_PANIC,
			"cbqunlock: cbq device attempting MODULE operation\n");
	}
}
