/*
 * Copyright (c) Sun Microsystems, Inc.  1995-1997. All rights reserved.
 *
 * License is granted to copy, to use, and to make and to use derivative works
 * for research and evaluation purposes, provided that Sun Microsystems is
 * acknowledged in all documentation pertaining to any such copy or
 * derivative work. Sun Microsystems grants no other licenses expressed or
 * implied. The Sun Microsystems  trade name should not be used in any
 * advertising without its written permission.
 *
 * SUN MICROSYSTEMS MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS
 * SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided "as is"
 * without express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this software.
 */
#ifndef _NETINET_CBQ_H
#define	_NETINET_CBQ_H

#pragma ident "@(#)cbq.h  1.5     97/03/12 SMI"

#include <sys/ioctl.h>
#include <sys/ioccom.h>

#ifdef __cplusplus
extern "C" {
#endif

#ifndef IFNAMSIZ
#define	IFNAMSIZ	16
#endif

#define	MAXQ		64	/* Max allowed queue length */

/*
 * Define filter structure
 */

#define	FILT_SIZE	48

typedef struct _filt cbq_filt_t;

#ifdef MASK_AND_MATCH
struct _filt {
	unsigned int	f_off;			/* Byte offset */
	int		f_len;			/* Length of SID  */
	u_char		f_mask[FILT_SIZE];	/* Mask of bits */
	u_char		f_val[FILT_SIZE];	/* Value of bits to compare */
};

#else

struct _filt {
	u_int		daddr;	/* Destination Address. */
	u_int		saddr;	/* Source Address */
	u_short		dport;	/* Destination Port. */
	u_short		sport;	/* Source port. */
	u_char		proto;	/* IP protocol Number. */
};
#endif

/*
 * Define a well known class handles
 */
#define	NULL_CLASS_HANDLE	0xffffffff
#define	ROOT_CLASS_HANDLE	0xfffffffe
#define	DEFAULT_CLASS_HANDLE	0xfffffffd
#define	CTL_CLASS_HANDLE	0xfffffffc

/*
 * Define structures associated with IOCTLS for cbq.
 */

/*
 * Define the CBQ interface structure.  This must be included in all
 * IOCTL's such that the CBQ driver may find the appropriate CBQ module
 * associated with the network interface to be affected.
 */
typedef struct cbq_interface {
	char	cbq_ifacename[IFNAMSIZ];
	u_int	cbq_ifacelen;
} cbq_iface_t;

/*
 * Define IOCTLs for CBQ.
 */
#define	CBQ_ENABLE		_IOW('Q', 1, struct cbq_interface)
#define	CBQ_DISABLE		_IOW('Q', 2, struct cbq_interface)

#define	CBQ_ADD_FILTER		_IOWR('Q', 3, struct cbq_add_filter)

struct cbq_add_filter {
	cbq_iface_t	cbq_iface;
	u_long		cbq_class_handle;
	cbq_filt_t	cbq_filter;

	u_long		cbq_filter_handle;
};

#define	CBQ_DEL_FILTER		_IOW('Q', 4, struct cbq_delete_filter)

struct cbq_delete_filter {
	cbq_iface_t	cbq_iface;
	u_long		cbq_filter_handle;
};

#define	CBQ_ADD_CLASS		_IOWR('Q', 5, struct cbq_add_class)

typedef struct cbq_class_spec {
	u_int		priority;
	u_int		nano_sec_per_byte;
	u_int		maxq;
	u_int		maxidle;
	int		minidle;
	u_int		offtime;
	u_long		parent_class_handle;
	u_long		borrow_class_handle;

	u_int		defaultclass;
	u_int		ctlclass;
	u_int		red;
	u_int		wrr;
	int		efficient;
} cbq_class_spec_t;

struct cbq_add_class {
	cbq_iface_t		cbq_iface;

	cbq_class_spec_t	cbq_class;
	u_long			cbq_class_handle;
};

#define	CBQ_DEL_CLASS		_IOW('Q', 6, struct cbq_delete_class)

struct cbq_delete_class {
	cbq_iface_t	cbq_iface;
	u_long		cbq_class_handle;
};

#define	CBQ_CLEAR_HIERARCHY	_IOW('Q', 7, struct cbq_interface)
#define	CBQ_ENABLE_STATS	_IOW('Q', 8, struct cbq_interface)
#define	CBQ_DISABLE_STATS	_IOW('Q', 9, struct cbq_interface)

#ifndef _CLASS_STATS
#define	_CLASS_STATS

typedef struct _class_stats_ {
	u_int		handle;
	u_int		depth;

	int		npackets;	/* packets sent in this class */
	int		nbytes;		/* bytes sent in this class */
	int		over;		/* # times went over limit */
	int		borrows;	/* # times tried to borrow */
	int		drops;		/* # times dropped packets */
	int		overactions;	/* # times invoked overlimit action */
	int		delays;		/* # times invoked delay actions */
} class_stats_t;
#endif

struct cbq_getstats {
	cbq_iface_t	iface;
	int		nclasses;
	class_stats_t	*stats;
};

#define	CBQ_GETSTATS		_IOW('Q', 10, struct cbq_getstats)


#ifdef _KERNEL

/*
 * Define macros only good for kernel drivers and modules.
 */
#define	N_MAX_CBQ_DEVS	1
#define	CBQ_DEVICE	0x1
#define	CBQ_MODULE	0x2

#define	ENABLE		0x01
#define	DISABLE		0x00

#define	IS_FAST_PATH		0x01
#define	IS_NOT_FAST_PATH	0x00

#define	CBQ_HI_WATER		(MAXQ * 1024)
#define	CBQ_LO_WATER		((MAXQ / 4) * 1024)

#define	CBQ_WATCHDOG    	(HZ / 20)
#define	CBQ_TIMEOUT		1
#define	CBQ_LS_TIMEOUT		drv_usectohz(20000)
#define	CBQ_VIRTUAL_IF_THRESH	1000
#define	CBQ_ACTIVE		1
#define	CBQ_INACTIVE		0

#define	CBQ_MAX_CLASSES		256
#define	CBQ_MAX_FILTERS		256

/*
 * Define State structures.
 */

typedef struct _k_filter_cb_ {
	struct rm_class		*clp;
	cbq_filt_t		filter;

	u_long			handle;

	struct _k_filter_cb_	*next;
	struct _k_filter_cb_	*prev;
	int			matches;
} k_filter_cb_t;

typedef struct cbqstate {
	kmutex_t		cbq_lock;
	kcondvar_t		cv;
#ifdef DEBUG
	int			cbqstarts;
	int			cbqfrees;
#endif
	queue_t			*cbq_q;
	mblk_t			*cbq_freeit;

	int			mac_addr_sz;
	int			is_pktsched_inited;
	int			is_pktsched_destroyed;
	int			use_esballoc;
	int			cbq_timeout_id;
	int			cbq_tl_timeout_id;
	int			cbq_busy;
	int			cbq_enabled;
	int			qlen;
	int			q_full;
	u_long			threshold;

	frtn_t			cbq_retn;

	struct rm_class		**cbq_class_tbl;

#ifndef MASK_AND_MATCH
#define	CBQ_FILTER_TBL_SZ	256
#define	CBQ_FILTER_MASK		(CBQ_FILTER_TBL_SZ - 1)
	k_filter_cb_t		*filters[CBQ_FILTER_TBL_SZ];
#else
	k_filter_cb_t		*filters;
	u_long			k_filt_handles;
#endif
	int			cbq_stats_timeout_id;
	int			cbq_stats_enabled;
	int			cbq_filter_debug;

	struct rm_ifdat		ifnp;
} cbq_mod_state_t;

typedef struct cbq_ill_s {
	struct cbq_ill_s	*cbq_ill_next;

	char			cbq_ill_name[IFNAMSIZ];
	int			cbq_ill_name_length;
	int			cbq_ill_ppa;

	struct cbqstate		cbq_state;
} cbq_ill_t;

#ifndef MASK_AND_MATCH
#define	CBQ_GET_HASH_INDEX(addr) \
	(((addr) + (addr >> 8)) & (CBQ_FILTER_TBL_SZ - 1))
#define	CBQ_GET_HINDEX(handle) ((handle) >> 24)
#endif

typedef struct cbq_ds_s {
	int		cbq_open;
	queue_t		*cbq_q;
} cbq_ds_t;

typedef struct cbq_s {
	int			cbq_id;
	struct cbq_s		*cbq_next;
	union {
		cbq_ill_t	cbq_ms;
		cbq_ds_t	cbq_ds;
	} cbq;
} cbq_t;

void		cbqenque(queue_t *, mblk_t *, int);
void		cbqstart(queue_t *, mblk_t *, cbq_mod_state_t *);
void		cbqfree(void *);
void		cbqwatchdog(queue_t *);
void		cbqlock(queue_t *);
void		cbqunlock(queue_t *);

int		cbq_class_destroy(cbq_mod_state_t *, struct rm_class *);
struct rm_class *cbq_class_create(cbq_mod_state_t *, struct cbq_add_class *,
			u_long, struct rm_class *, struct rm_class *);
int		cbq_class_queue(queue_t *, mblk_t *, int);

#ifdef CBQ_DEBUG
#define	ENTRY(x)	cmn_err(CE_CONT, "%s: entering\n", x)
#define	EXIT(x)		cmn_err(CE_CONT, "%s: exiting\n", x)
#else CBQ_DEBUG
#define	ENTRY(x)
#define	EXIT(x)
#endif /* CBQ_DEBUG */

#endif /* _KERNEL */

#ifdef __cplusplus
}
#endif

int cbq_debug(void);

#endif /* !_NETINET_CBQ_H */
