/*
 * Copyright (c) Sun Microsystems, Inc.  1993-1997. All rights reserved.
 *
 * License is granted to copy, to use, and to make and to use derivative works
 * for research and evaluation purposes, provided that Sun Microsystems is
 * acknowledged in all documentation pertaining to any such copy or
 * derivative work. Sun Microsystems grants no other licenses expressed or
 * implied. The Sun Microsystems  trade name should not be used in any
 * advertising without its written permission.
 *
 * SUN MICROSYSTEMS MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS
 * SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided "as is"
 * without express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this software.
 */

#ident "@(#)cbq_class.c  1.6     97/03/12 SMI"

/*
 * Copyright (c) 1993-1997 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and the Network Research Group at
 *	Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

static char sccsid[] = "@(#) cbq_class.c 1.6@(#)";
#pragma ident "@(#) cbq_class.c 1.6@(#)"
#ident "@(#) cbq_class.c 1.6@(#)"

#include <sys/types.h>
#include <sys/param.h>
#include <sys/sysmacros.h>
#include <sys/stream.h>
#include <sys/strsun.h>
#include <sys/stropts.h>
#include <sys/signal.h>
#include <sys/user.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/devops.h>
#include <sys/stat.h>
#include <sys/ethernet.h>
#include <sys/kmem.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <inet/common.h>
#include <inet/ip.h>

#include <sys/ddi.h>
#include <sys/sunddi.h>

#include <netinet/rm_class.h>
#include <netinet/cbq.h>

/*
 * Local functions.
 */
static caddr_t	ret_off(mblk_t *, u_int);

/*
 * Local data types.
 */
struct ip {
#ifdef _BIT_FIELDS_LTOH
	u_char  ip_hl:4,	/* header length */
	ip_v:4;			/* version */
#else
	u_char  ip_v:4,		/* version */
		ip_hl:4;	/* header length */
#endif
	u_char  ip_tos;		/* type of service */
	short   ip_len;		/* total length */
	u_short ip_id;		/* identification */
	short   ip_off;		/* fragment offset field */
#define	IP_DF 0x4000		/* dont fragment flag */
#define	IP_MF 0x2000		/* more fragments flag */
	u_char  ip_ttl;		/* time to live */
	u_char  ip_p;		/* protocol */
	u_short ip_sum;		/* checksum */
	struct  in_addr ip_src;
	struct  in_addr	ip_dst; /* source and dest address */
};

/*
 * int
 * cbq_class_destroy(cbq_mod_state_t *, struct rm_class *) - This
 *	function destroys a given traffic class.  Before destorying
 *	the class, all traffic for that class is released.
 */

int
cbq_class_destroy(cbq_mod_state_t *cbqp, struct rm_class *classp)
{
	rmc_delete_class(&cbqp->ifnp, classp);
	return (0);
}

/*
 * struct rm_class *
 * cbq_class_create(cbq_mod_state_t *cbqp, struct cbq_add_class *acp,
 *		u_long handle, struct rm_class *parent,
 *		struct rm_class *borrow)
*
 * This function create a new traffic class in the CBQ class hierarchy of
 * given paramters.  The class that created is either the root, default,
 * or a new dynamic class.  If CBQ is not initilaized, the the root class
 * will be created.
 */

struct rm_class *
cbq_class_create(cbq_mod_state_t *cbqp, struct cbq_add_class *acp,
	u_long handle, struct rm_class *parent, struct rm_class *borrow)
{
	struct rm_class		*cl;
	cbq_class_spec_t	*spec = &acp->cbq_class;

	if (!parent) {
		if (cbqp->ifnp.root_) {
			cmn_err(CE_CONT,
				"CBQ_ADD_CLASS: Root Class exists\n");
			return (NULL);
		}
		rmc_init(cbqp->cbq_q, &cbqp->ifnp, spec->nano_sec_per_byte,
			cbqwatchdog, MAXQ, RM_MAXQUEUED,
			spec->maxidle, spec->minidle, spec->offtime, spec->wrr);
		cbqp->ifnp.efficient_ = spec->efficient;
		return (cbqp->ifnp.root_);
	} else if (acp->cbq_class.defaultclass) {
		if (cbqp->ifnp.default_) {
			cmn_err(CE_CONT,
				"CBQ_ADD_CLASS: Default Class exists\n");
			return (NULL);
		}

		cbqp->ifnp.default_ = rmc_newclass(spec->priority,
			&cbqp->ifnp, spec->nano_sec_per_byte,
				rmc_delay_action, MAXQ, parent, borrow,
				spec->maxidle, spec->minidle, spec->offtime);

		return (cbqp->ifnp.default_);
	} else if (acp->cbq_class.ctlclass) {
		cbqp->ifnp.ctl_ = rmc_newclass(spec->priority,
			&cbqp->ifnp, spec->nano_sec_per_byte,
				rmc_delay_action, MAXQ, parent, borrow,
				spec->maxidle, spec->minidle, spec->offtime);

		return (cbqp->ifnp.ctl_);
	} else {
		cl = rmc_newclass(spec->priority, &cbqp->ifnp,
			spec->nano_sec_per_byte,
			rmc_delay_action, MAXQ, parent, borrow,
			spec->maxidle, spec->minidle, spec->offtime);

		if (!cl)
			return ((struct rm_class *)NULL);

		cbqp->cbq_class_tbl[handle] = cl;
		return (cl);
	}
}

/*
 * void
 * cbq_class_queue(queue_t *q, mblk_t *mp, int fastpath) - This function
 *	attempts to classify the packet into a traffic class for CBQ
 *	manage.  If the packet is a RSVP, IGMP, or RIP message, just forward
 *	it.  Otherwise, classify and return.
 */

int
cbq_class_queue(queue_t *q, mblk_t *mp, int fastpath)
{
	cbq_t		*cqp = (cbq_t *)q->q_ptr;
	struct cbqstate *cbqp = (struct cbqstate *)&cqp->cbq.cbq_ms.cbq_state;
	k_filter_cb_t 	*kfp;

	int		i;
	caddr_t		datap, paddr;
	u_long		offset;
	u_int		pseudo_hdr_sz;
	mblk_t		*nmp, *bp;

	u_int		daddr, saddr;
	u_int		hindex;
	int		match;
	unsigned	iphdrlen;
	struct _ports_ {
		u_short sport;
		u_short dport;
	} *ports;

	/*
	 * Add Protocol macros.
	 */
#ifndef IPPROTO_RSVP
#define IPPROTO_RSVP	46
#endif

#ifndef IPPROTO_PIM
#define IPPROTO_PIM	103
#endif

	/*
	 * Fastpatch means the device below supports the FAST path
	 * ioctl and that the pseudo header size is known before hand.
	 */
	if (fastpath) {
		if (MBLKL(mp) == cbqp->mac_addr_sz) {
			pseudo_hdr_sz = 0;
			bp = mp->b_cont;
		} else {
			pseudo_hdr_sz = cbqp->mac_addr_sz;
			bp = mp;
		}
	} else {
		pseudo_hdr_sz = 0;
		bp = mp->b_cont;
	}

	/*
	 * Point to the IP header.
	 */
	datap = ret_off(bp, pseudo_hdr_sz);

	/*
	 * Are we aligned, if so do work on aligned header.
	 */
	if ((datap != NULL) && OK_32PTR(datap)) {
		ipha_t *ip = (ipha_t *)datap;
		iphdrlen = IPH_HDR_LENGTH(ip);

		/*
		 * If we are a RSVP, IGMP, or RIP message put it out the
		 * interface immediately, otherwise classify the packet.
		 */
		switch (ip->ipha_protocol) {
		case IPPROTO_PIM:
			/*
			 * XXX - need to check for PIM-SM register packet, if
			 *	it is one need to look inside to the 
			 * 	encapsulated multicast packet.
			 */
		case IPPROTO_RSVP:
		case IPPROTO_IGMP:
			goto ip_ctl_class;

		case IPPROTO_UDP:
			paddr = datap + iphdrlen;
			ports = (struct _ports_ *)paddr;

			/*  RIP packet? */
			if (ports->dport == IPPORT_ROUTESERVER)
				goto ip_ctl_class;
			break;

		case IPPROTO_TCP:
			paddr = datap + iphdrlen;
			ports = (struct _ports_ *)paddr;
			break;

		case IPPROTO_ENCAP:
		default:
			/*
			 * XXX - IP encapsulated packets are just shipped out
			 *	as best effort.
			 */
			goto ip_default_class;
		}

		/*
		 * Loop through filter list to find match.  If no match
		 * found, packet will be queue to the default class.
		 */
		daddr = ip->ipha_dst;
		saddr = ip->ipha_src;

		kfp = cbqp->filters[CBQ_GET_HASH_INDEX(daddr)];
		while (kfp != NULL) {
			/*
			 * Check fields of IP header.
			 */
			match = 1;
			match &= (daddr == kfp->filter.daddr);
			if (kfp->filter.dport)
				match &= (ports->dport == kfp->filter.dport);
			if (kfp->filter.proto)
				match &= (ip->ipha_protocol ==
					kfp->filter.proto);
			if (kfp->filter.saddr)
				match &= (saddr == kfp->filter.saddr);
			if (kfp->filter.sport)
				match &= (ports->sport == kfp->filter.sport);

			if (match) {
				if (kfp->clp) {
					rmc_queue_packet(kfp->clp, mp);
					return (0);
				} else
					goto ip_default_class;
			}

			/*
			 * No match, get next filter and try again.
			 */
			kfp = kfp->next;
		}
	} else if (datap) {
		/*
		 * Non-aligned header case.
		 */
		struct ip *ip = (struct ip *)datap;
		iphdrlen = (ip->ip_hl * sizeof (unsigned int));

		/*
		 * If we are a RSVP, IGMP, or RIP message put it out the
		 * interface immediately, otherwise classify the packet.
		 */
		switch (ip->ip_p) {
		case IPPROTO_PIM:
			/*
			 * XXX - need to check for PIM-SM register packet, if
			 *	it is one need to look inside to the 
			 * 	encapsulated multicast packet.
			 */
		case IPPROTO_RSVP:
		case IPPROTO_IGMP:
			goto ip_ctl_class;

		case IPPROTO_UDP:
			paddr = datap + iphdrlen;
			ports = (struct _ports_ *)paddr;

			/*  RIP packet? */
			if (ports->dport == IPPORT_ROUTESERVER)
				goto ip_ctl_class;
			break;

		case IPPROTO_TCP:
			paddr = datap + iphdrlen;
			ports = (struct _ports_ *)paddr;
			break;

		case IPPROTO_ENCAP:
		default:
			/*
			 * XXX - IP encapsulated packets are just shipped out
			 *	as best effort.
			 */
			goto ip_default_class;
		}

		/*
		 * Loop through filter list to find match.  If no match
		 * found, packet will be queue to the default class.
		 */
		daddr = ip->ip_dst.s_addr;
		saddr = ip->ip_src.s_addr;

		kfp = cbqp->filters[CBQ_GET_HASH_INDEX(daddr)];
		while (kfp != NULL) {
			/*
			 * Check fields of IP header.
			 */
			match = 1;
			match &= (daddr == kfp->filter.daddr);
			if (kfp->filter.proto)
				match &= (ip->ip_p == kfp->filter.proto);
			if (kfp->filter.saddr)
				match &= (saddr == kfp->filter.saddr);
			if (kfp->filter.sport)
				match &= (ports->sport == kfp->filter.sport);
			if (kfp->filter.dport)
				match &= (ports->dport == kfp->filter.dport);

			if (match) {
				if (kfp->clp) {
					rmc_queue_packet(kfp->clp, mp);
					return (0);
				} else
					goto ip_default_class;
			}

			/*
			 * No match, get next filter and try again.
			 */
			kfp = kfp->next;
		}
	}

	/*
	 * No match in the filter list/table.  So, give packet to
	 * default class.
	 */
ip_default_class:
	if (cbqp->ifnp.default_)
		rmc_queue_packet(cbqp->ifnp.default_, mp);
	else
		freemsg(mp);
	return (0);

ip_ctl_class:
	if (cbqp->ifnp.ctl_)
		rmc_queue_packet(cbqp->ifnp.ctl_, mp);
	else
		freemsg(mp);
	return (0);
}

static caddr_t
ret_off(mblk_t *mp, u_int offset)
{
	if (mp)
		return ((caddr_t)mp->b_rptr + offset);
	else
		return ((caddr_t)NULL);
}
