/* 
 * Copyright (c) 1996 by Sun Microsystems, Inc.
 */

#ident "@(#)cbq_config.c  1.17     97/04/28 SMI"

/* 
 * Copyright (c) Sun Microsystems, Inc.  1994. All rights reserved. 
 *
 * License is granted to copy, to use, and to make and to use derivative 
 * works for research and evaluation purposes, provided that Sun Microsystems is
 * acknowledged in all documentation pertaining to any such copy or derivative
 * work. Sun Microsystems grants no other licenses expressed or implied. The
 * Sun Microsystems  trade name should not be used in any advertising without
 * its written permission.
 *
 * SUN MICROSYSTEMS MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF
 * THIS SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided "as is"
 * without express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this software.
 */

#include <stdio.h>
#include <strings.h>
#include <cbqd.h>

#define EQUAL(s1, s2)	(strcmp((char *)(s1), (char *)(s2)) == 0)
#define ADDR_SZ	8
#define PORT_SZ 4
#define PROTO_SZ 2

/*
 * Forward declarations.
 */
static u_char  *next_word();
static void     gethexstring(u_char * s, u_char * buf, int *plen, u_int max);

/*
 * Read the config file to learn about tunnel vifs and non-default phyint
 * parameters.
 */

int
cbq_do_config(char *configfilename)
{
	FILE           *f;
	u_char          linebuf[200], *s, *w;
	u_char		nbuf[20];
	int		len;
	int             line_no = 0;
	int             ret = 0;
	int		nclasses = 0;

	char		*ifname;
	long            bandwidth;
	int             is_shared;
	int		is_sched;
	int             discipline;
#define	DISCIPLINE_NONE		0
#define	DISCIPLINE_CBQ		1
#define DISCIPLINE_CSZ		2
#define DISCIPLINE_CTLLOAD	3

	u_char         *clname;
	u_char         *parent;
	u_char         *borrow;
	u_int           pri;
	u_int		efficient;
	u_int		is_stat_enabled;
	u_int		is_wrr;
	u_int           is_default;
	u_int           pbandwidth;
	u_int           maxdelay;
	u_int           maxburst;
	u_int           minburst;
	u_int           av_pkt_size;
	u_int           max_pkt_size;
	cbq_tos_t        admission_type;
	cbq_ol_t        overlimit_action;
	cbq_filt_t	sfilt;
	u_int		protocol;
	
	f = fopen(configfilename, "r");
	if (f == NULL) {
		fprintf(stderr, "Can't open %s", configfilename, 0);
		return (0);
	}

	while (fgets((char *) linebuf, sizeof(linebuf), f) != NULL) {
		s = linebuf;
		line_no++;
		if (EQUAL((w = next_word(&s)), "")) {
			/*
			 * blank or comment line; ignore
			 */
		} else if (EQUAL(w, "interface")) {
			bandwidth = 0;
			efficient = 0;
			is_stat_enabled = 0;
			is_shared = 0;
			is_sched = 1;
			is_wrr = 0;
			discipline = DISCIPLINE_NONE;

			/*
			 * Get interface name.
			 */
			if (EQUAL((w = next_word(&s)), "")) {
				fprintf(stderr, 
				    "Missing interface name in %s, line %d",
				    configfilename, line_no);
				goto fatal_err;
			}
			ifname = (char *)w;

			/*
			 * Look for "shared", "bandwidth" & scheduling
			 * discipline options.
			 */
			while (!EQUAL((w = next_word(&s)), "")) {
				if (EQUAL(w, "bandwidth")) {
					bandwidth = atol((char*)next_word(&s));
				} else if (EQUAL(w, "efficient")) {
					efficient = 1;
				} else if (EQUAL(w, "shared")) {
					is_shared = 1;
				} else if (EQUAL(w, "stats")) {
					is_stat_enabled = 1;
				} else if (EQUAL(w, "cbq")) {
					discipline = DISCIPLINE_CBQ;
				} else if (EQUAL(w, "cbq-wrr")) {
					discipline = DISCIPLINE_CBQ;
					is_wrr = 1;
				} else if (EQUAL(w, "nosched")) {
					is_sched = 0;
				} else if (EQUAL(w, "csz")) {
					fprintf(stderr, 
					    "'csz' not yet supported - %s, line %d",
					    configfilename, line_no);
					goto fatal_err;	/* just ignore for now */
				} else {
					fprintf(stderr, 
					    "Unknown keyword '%s' in %s, line %d",
					    w, configfilename, line_no);
					goto fatal_err;	/* just ignore unknown
							 * keywords for now */
				}
			}
			if (discipline == DISCIPLINE_CBQ) {
				/* Add this to list of CBQ interfaces */
				if (cbq_add_if(ifname, bandwidth,
						is_shared, is_stat_enabled,
						is_sched, is_wrr, efficient)) {
					 fprintf(stderr,
					 "can't init cbq on interface '%s'",
					    ifname);
					goto fatal_err;	/* Make fatal */
				}
			}
			if (!EQUAL(w, ""))
				continue;
		} else if (EQUAL(w, "class")) {
			parent = NULL;
			borrow = NULL;
			pri = 1;
			is_default = 0;
			pbandwidth = 100;
			maxdelay = 100;
			maxburst = 20;
			minburst = maxburst / 10;
			av_pkt_size = 1000;
			max_pkt_size = 1500;
			admission_type = CBQ_QOS_NONE;
			overlimit_action = OL_DELAY;

			/*
			 * Get scheduling class.
			 */
			if (EQUAL((w = next_word(&s)), "")) {
				fprintf(stderr,
				    "missing scheduling discipline in %s, line %d",
				    configfilename, line_no);
				goto fatal_err;
			}
			if (!EQUAL(w, "cbq")) {
				fprintf(stderr,
				    "unknown scheduling discipline '%s' in %s, line %d",
				    w, configfilename, line_no);
				goto fatal_err;
			}
			/*
			 * Get interface name.
			 */
			if (EQUAL((w = next_word(&s)), "")) {
				fprintf(stderr,
				    "missing interface name in %s, line %d",
				    configfilename, line_no);
				goto fatal_err;
			}
			ifname = (char *)w;

			/*
			 * Get class name.
			 */
			if (EQUAL((w = next_word(&s)), "")) {
				fprintf(stderr, 
				    "missing class name in %s, line %d",
				    configfilename, line_no);
				goto fatal_err;
			}
			clname = w;

			/*
			 * Get parent class.
			 */
			if (EQUAL((w = next_word(&s)), "")) {
				fprintf(stderr,
				    "missing parent class in %s, line %d",
				    configfilename, line_no);
				goto fatal_err;
			}
			if (!EQUAL(w, "null")) {
				parent = w;
			}
			/*
			 * Look for options.
			 */
			while (!EQUAL((w = next_word(&s)), "")) {
				if (EQUAL(w, "priority")) {
					pri = (u_int)atoi((char*)next_word(&s));
				} else if (EQUAL(w, "default")) {
					is_default = 1;
				} else if (EQUAL(w, "admission")) {
					w = next_word(&s);
					if (EQUAL(w, "guaranteed"))
						admission_type =
							CBQ_QOS_GUARANTEED;
					else if (EQUAL(w, "predictive"))
						admission_type =
							CBQ_QOS_PREDICTIVE;
					else if (EQUAL(w, "cntlload"))
						admission_type =
							CBQ_QOS_CNTR_LOAD;
					else if (EQUAL(w, "cntldelay"))
						admission_type =
							CBQ_QOS_CNTR_DELAY;
					else if (EQUAL(w, "none"))
						admission_type = CBQ_QOS_NONE;
					else {
						fprintf(stderr,
						    "unknown admission type - %s, line %d",
						    w, line_no);
						goto fatal_err;
					}
				} else if (EQUAL(w, "maxdelay")) {
					maxdelay = (u_int)
						atoi((char*)next_word(&s));
				} else if (EQUAL(w, "borrow")) {
					borrow = next_word(&s);
				} else if (EQUAL(w, "pbandwidth")) {
					pbandwidth = (u_int)
						atoi((char*)next_word(&s));
				} else if (EQUAL(w, "maxburst")) {
					maxburst = (u_int)
						atoi((char*)next_word(&s));
				} else if (EQUAL(w, "minburst")) {
					minburst = (u_int)
						atoi((char*)next_word(&s));
				} else if (EQUAL(w, "packetsize")) {
					av_pkt_size = (u_int)
						atoi((char*)next_word(&s));
				} else {
					fprintf(stderr, 
					    "Unknown keyword '%s' in %s, line %d",
					    w, configfilename, line_no);
					goto fatal_err;	/* just ignore unknown
							 * keywords for now */
				}
			}

			ret = cbq_add_static_class(ifname, (char *)clname,
				(char *)parent, (char *)borrow, pri,
				is_default, pbandwidth, maxdelay,
				maxburst, minburst, av_pkt_size, max_pkt_size,
				admission_type, overlimit_action);

			if (ret) {
				fprintf(stderr,
				    "can't add class '%s' on interface '%s'",
				    clname, ifname);
				goto fatal_err;	/* Make fatal */
			} else
				nclasses++;

			if (!EQUAL(w, ""))
				continue;
		} else if (EQUAL(w, "filter")) {
			u_int	sport, dport, proto;

			/*
			 * Get interface name.
			 */
			if (EQUAL((w = next_word(&s)), "")) {
				fprintf(stderr,
				    "missing interface name in %s, line %d",
				    configfilename, line_no);
				goto fatal_err;
			}
			ifname = (char *)w;

			/*
			 * Get class name.
			 */
			if (EQUAL((w = next_word(&s)), "")) {
				fprintf(stderr,
				    "missing class name in %s, line %d",
				    configfilename, line_no);
				goto fatal_err;
			}
			clname = w;

			/*
			 * Get Destination Address.
			 */
			if (EQUAL((w = next_word(&s)), "")) {
				fprintf(stderr,
				    "missing filter value in %s, line %d",
				    configfilename, line_no);
				goto fatal_err;
			}
			sscanf((const char *)w, "%x", &sfilt.daddr);

			/*
			 * Get Destination port.
			 */
			if (EQUAL((w = next_word(&s)), "")) {
				fprintf(stderr, 
				    "missing filter mask in %s, line %d",
				    configfilename, line_no);
				goto fatal_err;
			}
			sscanf((const char *)w, "%x", &dport);
			sfilt.dport = (u_short)dport;

			/*
			 * Get Source Address.
			 */
			if (EQUAL((w = next_word(&s)), "")) {
				fprintf(stderr, 
				    "missing filter value in %s, line %d",
				    configfilename, line_no);
				goto fatal_err;
			}
			sscanf((const char *)w, "%x", &sfilt.saddr);

			/*
			 * Get Source port.
			 */
			if (EQUAL((w = next_word(&s)), "")) {
				fprintf(stderr,
				    "missing filter mask in %s, line %d",
				    configfilename, line_no);
				goto fatal_err;
			}
			sscanf((const char *)w, "%x", &sport);
			sfilt.sport = (u_short)sport;

			/*
			 * Get Protocol
			 */
			if (EQUAL((w = next_word(&s)), "")) {
				fprintf(stderr, 
				    "missing filter mask in %s, line %d",
				    configfilename, line_no);
				goto fatal_err;
			}
			sscanf((const char *)w, "%x", &proto);
			sfilt.proto = (u_char)proto;

			if (ret = cbq_add_static_filter(ifname,
					(char *)clname, &sfilt)) {
				fprintf(stderr,
				    "can't add filter to class '%s' on interface '%s'",
				    clname, ifname);
				goto fatal_err;
			}
			if (!EQUAL(w, ""))
				continue;
		} else {
			fprintf(stderr, 
			    "unknown command '%s' in %s", w, configfilename);
			goto fatal_err;
		}
	}

	if (nclasses)
		nclasses++;
	fclose(f);
	return (nclasses);

fatal_err:
	cbq_destroyall();
	fclose(f);
	return (0);
}


/*
 * Return a pointer to the next "word" in the string to which '*s' points,
 * lower-cased and null terminated, and advance '*s' to point beyond the
 * word. Words are separated by blanks and/or tabs, and the input string is
 * considered to terminate at a newline, '#' (comment), or null character. If
 * no words remain, a pointer to a null string ("") is returned.  Warning:
 * This function clobbers the input string.
 */
static u_char  *
next_word(s)
	u_char        **s;
{
	u_char         *w;

	w = *s;
	while (*w == ' ' || *w == '\t')
		++w;

	*s = w;
	for (;;) {
		switch (**s) {

		case ' ':
		case '\t':
			**s = '\0';
			++*s;
			return (w);

		case '\n':
		case '#':
			**s = '\0';
			return (w);

		case '\0':
			return (w);

		default:
			if (isascii(**s) && isupper(**s))
				**s = tolower(**s);
			++*s;
		}
	}
}


static void
gethexstring(u_char *s, u_char *buf, int *plen, u_int max)
{
	register int    hi = 0, nibble;

	*plen = 0;
	memset(buf, 0, max);
	max = max << 1;
	while ((*plen <= max) && isxdigit(*s)) {
		nibble = toupper(*s) - '0';
		if (nibble > 9)
			nibble -= 7;
		if (hi) {
			*buf = *buf << 4 | nibble;
			buf++;
		} else
			*buf |= nibble;
		hi = !hi;
		(*plen)++;
		s++;
	}
	*plen = (*plen + 1) / 2;
}
