/* 
 * Copyright (c) 1996 by Sun Microsystems, Inc.
 */

#ident "@(#)cbqd.c  1.20     97/04/22 SMI"

#include <stdio.h>
#include <signal.h>
#include <thread.h>
#include <cbqd.h>

/*
 * Local macros.
 */

#define CMD_LINE_OPTS	"d:f:c:"
#define TV_DELTA(a, b, delta) { \
	register int    xxs;    \
\
	delta = (a)->tv_usec - (b)->tv_usec; \
	if ((xxs = (a)->tv_sec - (b)->tv_sec)) { \
		switch (xxs) { \
		default: \
			if (xxs < 0) { \
				fprintf(stderr, "cbqd: bogus timevals\n"); \
				exit(1); \
			} \
			delta = 0; \
			/* fall through */ \
		case 2: \
			delta += 1000000; \
			/* fall through */ \
		case 1: \
			delta += 1000000; \
			break; \
		} \
	} \
}

/*
 * Local variables.
 */
static int	debug = 0;
static char	*config = "cbqd.conf";
static int	done = 0;

/*
 * Local Functions.
 */
static void *sighandler(void *);

/*
 * Global Data.
 */
int		count = 0, nclasses = 0;
class_stats_t	*clstats = NULL;
class_stats_t	*pclstats = NULL;
float		*pstats = NULL;

int
main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;

	int		i, c, sig;
	int		maxcount = 10000;
	sigset_t	set;
	cbq_ifinfo_t	*ifinfo;
	thread_t	sigthreadid;
	int		inited = 0;
	int		diff;
	float		bytes_per_sec;

	struct timeval	last, now;

	while ((c = getopt(argc, argv, CMD_LINE_OPTS)) != EOF) {
		switch (c) {
		case 'd':
			debug = atoi(optarg);
			break;
		case 'f':
			config = optarg;
			break;
		case 'c':
			maxcount = atoi(optarg);
			break;
		}
	}

	if (thr_create((void *)NULL, 0, sighandler, (void *)NULL,
			THR_NEW_LWP | THR_BOUND, &sigthreadid)) {
		perror("sigthread");
		exit(1);
	}

	if (cbq_init(debug))
		goto err_exit;

	if ((nclasses = cbq_do_config(config)) == 0)
		goto err_exit;
	clstats = (class_stats_t *)malloc(sizeof(class_stats_t) * nclasses);
	pclstats = (class_stats_t *)malloc(sizeof(class_stats_t) * nclasses);
	pstats = (float *)malloc(sizeof (float) * nclasses);
	if (!clstats || !pclstats || !pstats)
		goto err_exit;
	
	if (cbq_enableall())
		goto err_exit;

	memset(pstats, 0, sizeof (float) * nclasses);
	memset(pclstats, 0, sizeof (class_stats_t) * nclasses);
	memset(clstats, 0, sizeof (class_stats_t) * nclasses);

	ifinfo = cbq_getifinfo("qe0");
	bytes_per_sec = (float)ifinfo->bandwidth / 8.0;

	gettimeofday(&last, 0);
	poll(0, 0, 1000);
	while (!done) {
		if (cbq_getstats("qe0", clstats, nclasses))
			done++;
		else {
			gettimeofday(&now, 0);
			TV_DELTA(&now, &last, diff);
			for (i = 0; i < nclasses; i++) {
				pstats[i] = clstats[i].nbytes -
					pclstats[i].nbytes;
				pstats[i] = pstats[i] * 1000000.0 /
					(bytes_per_sec * diff);
			}
			cbq_report_stats("qe0", clstats, pstats, nclasses);
			memcpy((void *)pclstats, (void *)clstats, 
				sizeof(class_stats_t) * nclasses);

			if (++count >= maxcount)
				done++;
			last = now;
			poll(0, 0, 1000);
		}
	}

err_exit:
	cbq_destroyall();
	exit(1);
}

static void *
sighandler(void *arg)
{
        sigset_t        set;

	sigemptyset(&set);
	sigaddset(&set, SIGTERM);
	sigaddset(&set, SIGHUP);

	switch (sigwait(&set)) {
	case SIGINT:
	case SIGTERM:
	case SIGHUP:
		done++;
		break;
	}

	thr_exit(0);
}
