/*
 * Copyright (c) 1996 by Sun Microsystems, Inc.
 */
#ifndef _CBQD_H
#define	_CBQD_H

#pragma ident "@(#)cbqd.h  1.11     97/04/28 SMI"

#ifdef __cplusplus
extern "C" {
#endif 

#include <sys/types.h>
#include <sys/ioctl.h>
#include <netinet/cbq.h>

/*
 * Local Data structures.
 */

typedef struct cbq_ifinfo {
	struct cbq_ifinfo	*next;
	char			*ifname;
	long			bandwidth;
	double			nsPerByte;
	int			is_stats_enabled;
	int			is_shared;
	int			is_wrr;
	int			is_efficient;
	int			use_cbq;
	int			use_sbm;
	struct cbq_classinfo	*cd_class;
	struct cbq_classinfo	*cllist;
} cbq_ifinfo_t;

typedef enum {
	CBQ_QOS_NONE, CBQ_QOS_GUARANTEED, CBQ_QOS_PREDICTIVE,
	CBQ_QOS_CNTR_DELAY, CBQ_QOS_CNTR_LOAD
} cbq_tos_t;

typedef struct cbq_acinfo {
	cbq_tos_t		type;
	u_int			pri;
	long			bandwidth;
	long			allocated;
	u_int			maxdelay;
	u_int			maxburst;
	u_int			av_pkt_size;
	u_int			nflows;
	struct cbq_flowinfo	*flow_list;
} cbq_acinfo_t;

typedef enum {
	OL_NONE, OL_DROP, OL_DELAY
} cbq_ol_t;

typedef struct cbq_classinfo {
	struct cbq_classinfo	*next;
	u_long			handle;
	char			*clname;
	cbq_ifinfo_t		*ifinfo;
	struct cbq_classinfo	*parent_cl;
	struct cbq_classinfo	*borrow_cl;
	cbq_acinfo_t		acinfo;
} cbq_classinfo_t;

typedef struct cbq_fltrinfo {
	struct cbq_fltrinfo	*next;
	u_long			handle;
	cbq_filt_t		filter;
	struct cbq_flowinfo	*flow;
} cbq_fltrinfo_t;


/* Per-service RSPEC definitions */
typedef struct {
	u_long		g_rate;	/* GUARANTEED: flow rate parameter	*/
} cbq_g_rspec_t;

typedef struct {
    ulong		p_dly;	/* PREDICTIVE: delay target parameter	*/
    				/* May also be used as priority..	*/
} cbq_p_rspec_t;

typedef union {
	cbq_g_rspec_t	rs_g;	/* TOS_ST_GUARANTEED			*/
	cbq_p_rspec_t	rs_p;	/* TOS_ST_PREDICTIVE			*/
} cbq_rspec_t;

/* Traffic spec (TSPEC) definitions */
typedef struct {
	u_long		tbf_bkt;	/* tbf bucket size: bits	*/
	u_long		tbf_rate;	/* tbf rate: bits/sec		*/
	u_long		tbf_buf;	/* tbf reshaping buffer size    */
					/* (currently MBZ)		*/
} cbq_tbf_t;

typedef struct {
	cbq_tbf_t	ts_tb;
} cbq_tspec_t;

typedef struct cbq_flowinfo {
	struct cbq_flowinfo	*next;
	cbq_classinfo_t		*clinfo;
	long			bandwidth; 
	u_long			handle;
	cbq_fltrinfo_t		*fltr_list;
} cbq_flowinfo_t;

typedef void	*cbq_handle_t;

/*
 * Cbq kernel Interface routines.
 */
int		cbq_init(int debug);
int		cbq_add_if(char *ifname, long bandwidth, int is_shared,
			int is_stats_enabled, int is_sched, int is_wrr,
			int efficient);
int		cbq_enableall(void);
int		cbq_enable(char *ifname);
int		cbq_disableall(void);
int		cbq_disable(char *ifname);
int		cbq_add_static_class(char *ifname, char *clname, char *parent,
			char *borrow, u_int pri, u_int is_default,
			u_int pbandwidth, u_int maxdelay,
			u_int maxburst, u_int minburst,
			u_int avg_pkt_sz, u_int max_pkt_sz,
			cbq_tos_t admission_type,
			cbq_ol_t overlimit_action);
int		cbq_add_static_filter(char *ifname, char *class,
			cbq_filt_t *filt);
cbq_fltrinfo_t	*cbq_set_filter(cbq_flowinfo_t *flinfo, cbq_filt_t *cfilter);
int		cbq_delfilter_all(cbq_flowinfo_t *flow_info);
int		cbq_delfilter(cbq_fltrinfo_t *fltr);
int		cbq_clear(char *ifname);
int		cbq_destroyall(void);
int		cbq_getstats(char *ifname, class_stats_t *clstats,
			int nclasses);
void		cbq_report_stats(char *interface,
			class_stats_t *stats, float *pstats, int nclasses);
cbq_acinfo_t	*cbq_getacinfo(char *ifname);
cbq_ifinfo_t	*cbq_getifinfo(char *ifname);



#ifdef __cplusplus
}
#endif 

#endif /* !_CBQD_H */
