/*
 * Copyright (c) Sun Microsystems, Inc.  1994-1997. All rights reserved.
 *
 * License is granted to copy, to use, and to make and to use derivative works
 * for research and evaluation purposes, provided that Sun Microsystems is
 * acknowledged in all documentation pertaining to any such copy or
 * derivative work. Sun Microsystems grants no other licenses expressed or
 * implied. The Sun Microsystems  trade name should not be used in any
 * advertising without its written permission.
 *
 * SUN MICROSYSTEMS MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS
 * SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided "as is"
 * without express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this software.
 */

#pragma ident	"@(#)cbqddi.c	1.3	97/03/24 SMI"

#ifndef BUILD_STATIC

#include <sys/types.h>
#include <sys/conf.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/stream.h>
#include <sys/errno.h>
#include <sys/modctl.h>
#include <sys/sunddi.h>
#include <sys/stat.h>
#include <sys/ddi.h>

#include <netinet/rm_class.h>
#include <netinet/cbq.h>

#define	D_SD_COMMENT "CBQ Streams Device v1.3"
#define	D_SD_INFO cbqinfo
#define	D_SD_NAME "cbq"
#define	D_SD_OPS_NAME cbq_d_ops
#define	D_SD_FLAGS D_MP|D_MTQPAIR|D_MTPUTSHARED|D_MTOUTPERIM|D_MTOCEXCL

/*
 * D_SD_FLAGS in the other tcp/ip modules have to these D_SD_FLAGS since they
 * are effectively clones of the ip driver with their module autopushed.
 */
#define	D_SM_COMMENT "CBQ Streams Module v1.3"
#define	D_SM_INFO cbqinfo
#define	D_SM_NAME "cbq"
#define	D_SM_OPS_NAME cbq_m_ops
#define	D_SM_FLAGS D_MP|D_MTQPAIR|D_MTPUTSHARED|D_MTOUTPERIM|D_MTOCEXCL

extern	int	nulldev();
extern	int	nodev();

extern	kmutex_t	cbq_g_lock;

#ifdef D_SD_INFO
extern	struct mod_ops mod_driverops;
extern	struct streamtab D_SD_INFO;

static int _mi_driver_info(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
			void **result);
static	int	_mi_driver_attach(dev_info_t * devi,  ddi_attach_cmd_t cmd);
static	int	_mi_driver_identify(dev_info_t * devi);

/* inter-module dependencies */
char _depends_on[] = "drv/cbq";

static struct cb_ops _mi_driver_ops = {
	nulldev,		/* cb_open */
	nulldev,		/* cb_close */
	nodev,			/* cb_strategy */
	nodev,			/* cb_print */
	nodev,			/* cb_dump */
	nodev,			/* cb_read */
	nodev,			/* cb_write */
	nodev,			/* cb_ioctl */
	nodev,			/* cb_devmap */
	nodev,			/* cb_mmap */
	nodev,			/* cb_segmap */
	nochpoll,		/* cb_chpoll */
	ddi_prop_op,		/* cb_prop_op */
	&D_SD_INFO,		/* cb_stream */
	(int)D_SD_FLAGS		/* cb_flag */
};

struct dev_ops D_SD_OPS_NAME = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	_mi_driver_info,	/* devo_getinfo */
	_mi_driver_identify,	/* devo_identify */
	nulldev,		/* devo_probe */
	_mi_driver_attach,	/* devo_attach */
	nodev,			/* devo_detach */
	nodev,			/* devo_reset */
	&_mi_driver_ops,	/* devo_cb_ops */
	(struct bus_ops *)NULL	/* devo_bus_ops */
};

static struct modldrv modldrv = {
	&mod_driverops,	 D_SD_COMMENT, &D_SD_OPS_NAME
};
#endif

#ifdef D_SM_INFO
extern	struct streamtab D_SM_INFO;
extern struct mod_ops mod_strmodops;

static int _mi_module_info(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
			void **result);
static	int	_mi_module_attach(dev_info_t *devi, ddi_attach_cmd_t cmd);
static	int	_mi_module_identify(dev_info_t *devi);

static struct cb_ops _mi_module_ops = {
	nulldev,		/* cb_open */
	nulldev,		/* cb_close */
	nodev,			/* cb_strategy */
	nodev,			/* cb_print */
	nodev,			/* cb_dump */
	nodev,			/* cb_read */
	nodev,			/* cb_write */
	nodev,			/* cb_ioctl */
	nodev,			/* cb_devmap */
	nodev,			/* cb_mmap */
	nodev,			/* cb_segmap */
	nochpoll,		/* cb_chpoll */
	ddi_prop_op,		/* cb_prop_op */
	&D_SM_INFO,		/* cb_stream */
	(int)D_SM_FLAGS		/* cb_flag */
};

struct dev_ops D_SM_OPS_NAME = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	_mi_module_info,	/* devo_getinfo */
	_mi_module_identify,	/* devo_identify */
	nulldev,		/* devo_probe */
	_mi_module_attach,	/* devo_attach */
	nodev,			/* devo_detach */
	nodev,			/* devo_reset */
	&_mi_module_ops,	/* devo_cb_ops */
	(struct bus_ops *)NULL	/* devo_bus_ops */
};


static struct fmodsw _mi_module_fmodsw = {
	D_SM_NAME, &D_SM_INFO, D_SM_FLAGS
};

static struct modlstrmod modlstrmod = {
	&mod_strmodops, D_SM_COMMENT, &_mi_module_fmodsw
};
#endif

static struct modlinkage modlinkage = {
	MODREV_1, (void *)&modlstrmod, (void *)&modldrv, NULL
};

int
_init()
{
	return (mod_install(&modlinkage));
}

int
_fini()
{
	return (mod_remove(&modlinkage));
}

_info(modinfop)
	struct modinfo *modinfop;
{
	return (mod_info(&modlinkage, modinfop));
}

#ifdef D_SD_INFO
static	dev_info_t	* _mi_driver_dev_info;

/* ARGSUSED */
static int
_mi_driver_attach(dev_info_t *devi, ddi_attach_cmd_t cmd)
{
	ENTRY("_mi_driver_attach");

	_mi_driver_dev_info = devi;
	if (ddi_create_minor_node(devi, D_SD_NAME, S_IFCHR,
			0, NULL, CLONE_DEV) == DDI_FAILURE) {
		EXIT("_mi_driver_attach");
		ddi_remove_minor_node(devi, NULL);
		return (-1);
	}

	mutex_init(&cbq_g_lock, "CBQ global lock", MUTEX_DEFAULT,
		(void *)NULL);

	EXIT("_mi_driver_attach");
	return (DDI_SUCCESS);
}

/* ARGSUSED */
static int
_mi_driver_info(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
	void **result)
{
	register dev_t dev = (dev_t) arg;
	register int error;

	ENTRY("_mi_driver_info");

	switch (infocmd) {
	case DDI_INFO_DEVT2DEVINFO:
		if (_mi_driver_dev_info == NULL) {
			error = DDI_FAILURE;
		} else {
			*result = (void *) _mi_driver_dev_info;
			error = DDI_SUCCESS;
		}
		break;
	case DDI_INFO_DEVT2INSTANCE:
		*result = (void *)0;
		error = DDI_SUCCESS;
		break;
	default:
		error = DDI_FAILURE;
	}

	EXIT("_mi_driver_info");
	return (error);
}

static int
_mi_driver_identify(dev_info_t *devi)
{
	ENTRY("_mi_driver_identify");

	if (strcmp((char *)ddi_get_name(devi), D_SD_NAME) == 0) {
		EXIT("_mi_driver_identify");
		return (DDI_IDENTIFIED);
	}

	EXIT("_mi_driver_identify");
	return (DDI_NOT_IDENTIFIED);
}
#endif

#ifdef D_SM_INFO
static	dev_info_t	* _mi_module_dev_info;

/* ARGSUSED */
static int
_mi_module_attach(dev_info_t *devi, ddi_attach_cmd_t cmd)
{
	ENTRY("_mi_module_attach");

	_mi_module_dev_info = devi;

	EXIT("_mi_module_attach");
	return (DDI_SUCCESS);
}

/* ARGSUSED */
static int
_mi_module_info(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
	void **result)
{
	register dev_t dev = (dev_t) arg;
	register int error;

	ENTRY("_mi_module_info");

	switch (infocmd) {
	case DDI_INFO_DEVT2DEVINFO:
		if (_mi_module_dev_info == NULL) {
			error = DDI_FAILURE;
		} else {
			*result = (void *) _mi_module_dev_info;
			error = DDI_SUCCESS;
		}
		break;
	case DDI_INFO_DEVT2INSTANCE:
		*result = (void *)0;
		error = DDI_SUCCESS;
		break;
	default:
		error = DDI_FAILURE;
	}

	EXIT("_mi_module_info");
	return (error);
}

static int
_mi_module_identify (devi)
	dev_info_t * devi;
{
	ENTRY("_mi_module_identify");

	if (strcmp((char *)ddi_get_name(devi), D_SM_NAME) == 0) {
		EXIT("_mi_module_identify");
		return (DDI_IDENTIFIED);
	}

	EXIT("_mi_module_identify");
	return (DDI_NOT_IDENTIFIED);
}
#endif

#endif BUILD_STATIC
