/*
 * Copyright (c) Sun Microsystems, Inc. 1996-1997. All rights reserved.
 *
 * License is granted to copy, to use, and to make and to use derivative works
 * for research and evaluation purposes, provided that Sun Microsystems is
 * acknowledged in all documentation pertaining to any such copy or
 * derivative work. Sun Microsystems grants no other licenses expressed or
 * implied. The Sun Microsystems  trade name should not be used in any
 * advertising without its written permission.
 *
 * SUN MICROSYSTEMS MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS
 * SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided "as is"
 * without express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this software.
 */

#ident "@(#)pktsched.c  1.2     97/02/17 SMI"

#include <sys/types.h>
#include <sys/conf.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/stream.h>
#include <sys/strsun.h>
#include <sys/errno.h>
#include <sys/modctl.h>
#include <sys/sunddi.h>
#include <sys/stat.h>
#include <sys/ddi.h>
#include <sys/dlpi.h>
#include <sys/stropts.h>
#include <sys/strlog.h>
#include <sys/tihdr.h>
#include <sys/time.h>
#include <sys/tiuser.h>
#include <sys/kstat.h>
#include <sys/kmem.h>

#include <sys/socket.h>
#include <net/route.h>
#include <sys/sockio.h>
#include <netinet/in.h>
#include <inet/common.h>
#include <inet/mi.h>
#include <inet/ip.h>
#include <inet/ip_multi.h>
#include <inet/ip_ire.h>
#include <netinet/ip_mroute.h>
#include <sys/ethernet.h>

#include <netinet/pktsched.h>

int
pktsched_init(queue_t *q, pkt_sched_cb_t *cb, int *flag,
	kmutex_t *mtx, kcondvar_t *cv)
{
	mblk_t		*mp;
	struct iocblk	*iocp;

	if ((mp = allocb(sizeof (struct iocblk), BPRI_HI)) == NULL)
		return (0);
	if ((mp->b_cont = allocb(sizeof (pkt_sched_cb_t), BPRI_HI)) == NULL)
		return (0);

	/*
	 * Set up the IOCTL to the driver.
	 */
	mp->b_datap->db_type = M_IOCTL;
	bzero((char *)mp->b_wptr, sizeof (*iocp));
	iocp = (struct iocblk *)mp->b_wptr;
	iocp->ioc_cmd = DL_IOC_PKTSCHED_INIT;
	iocp->ioc_count = msgdsize(mp->b_cont);
	mp->b_wptr += sizeof (*iocp);

	bcopy((void *)cb, (void *)mp->b_cont->b_wptr, sizeof (pkt_sched_cb_t));
	mp->b_cont->b_wptr += sizeof (pkt_sched_cb_t);

	/*
	 * Send the IOCTL.
	 */
	putq(WR(q), mp);

	/*
	 * Wait for response.
	 */
	while (*flag == 0)
		cv_wait(cv, mtx);

	if (*flag == -1)
		return (0);
	return (1);
}

int
pktsched_teardown(queue_t *q, int *flag, kmutex_t *mtx, kcondvar_t *cv)
{
	mblk_t		*mp;
	struct iocblk	*iocp;

	if ((mp = allocb(sizeof (struct iocblk), BPRI_HI)) == NULL)
		return (0);

	/*
	 * Build the ioctl.
	 */
	mp->b_datap->db_type = M_IOCTL;
	bzero((char *)mp->b_wptr, sizeof (*iocp));
	iocp = (struct iocblk *)mp->b_wptr;
	iocp->ioc_cmd = DL_IOC_PKTSCHED_TEARDOWN;
	iocp->ioc_count = 0;
	mp->b_wptr += sizeof (*iocp);

	/*
	 * Send the IOCTL.
	 */
	putq(WR(q), mp);

	/*
	 * Wait for response.
	 */
	while (*flag == 0)
		cv_wait(cv, mtx);

	if (*flag == -1)
		return (0);
	return (1);
}
