/*
 * Copyright (c) Sun Microsystems, Inc.  1993-1997. All rights reserved.
 *
 * License is granted to copy, to use, and to make and to use derivative works
 * for research and evaluation purposes, provided that Sun Microsystems is
 * acknowledged in all documentation pertaining to any such copy or
 * derivative work. Sun Microsystems grants no other licenses expressed or
 * implied. The Sun Microsystems  trade name should not be used in any
 * advertising without its written permission.
 *
 * SUN MICROSYSTEMS MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS
 * SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided "as is"
 * without express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this software.
 */

/*
 * Copyright (c) 1991-1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Network Research
 *      Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#)rm_class.c  1.26     97/04/28 SMI"

#include <sys/param.h>
#ifndef	SOLARIS
#include <sys/mbuf.h>
#endif
#include <sys/buf.h>
#include <sys/map.h>
#include <sys/socket.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/debug.h>

#include <sys/time.h>
#ifndef	SOLARIS
#include <sys/kernel.h>
#endif

#include <net/if.h>
#include <net/if_arp.h>
#ifndef	SOLARIS
#include <net/netisr.h>
#endif
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#ifdef	SOLARIS
#include <sys/kmem.h>
#include <sys/stream.h>
#endif

#include <netinet/rm_class.h>
#include <netinet/rm_class_debug.h>

#ifdef	SOLARIS
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <netinet/cbq.h>
#endif

#ifdef SOLARIS
#define	panic(str)	cmn_err(CE_PANIC, (str));
#endif

/*
 * Local Macros
 */

#define reset_cutoff(ifd) { \
	ifd->cutoff_ = RM_MAXDEPTH; \
}

/*
 * Local routines.
 */

#ifdef USE_HRTIME
static int	rmc_satisfied(struct rm_class *cl, hrtime_t *now);
#else
static int	rmc_satisfied(struct rm_class *cl, struct timeval *now);
#endif
static void	rmc_wrr_set_weights(struct rm_ifdat *ifd);
static void	rmc_depth_compute(struct rm_class *cl);

static void	_rmc_addq(class_queue_t *q, mbuf_t *m);
static void	_rmc_dropq(class_queue_t *q);
static mbuf_t	*_rmc_getq(class_queue_t *q);

/*
 * local data.
 */
#if defined(CBQTRACE)
struct cbqtrace		cbqtrace_buffer[NCBQTRACE+1];
struct cbqtrace		*cbqtrace_ptr = NULL;
int			cbqtrace_count;
#endif

/*
 * Table for mapping a bit mask (with one bit per priority level)
 * into the number of the highest priority set bit.
 */
u_char rmc_mask2pri[] = {
	0, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, /* 00 - 0f */
	4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 10 - 1f */
	5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 20 - 2f */
	5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 30 - 3f */
	6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 40 - 4f */
	6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 50 - 5f */
	6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 60 - 6f */
	6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 70 - 7f */
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 80 - 8f */
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 90 - 9f */
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* a0 - af */
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* b0 - bf */
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* c0 - cf */
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* d0 - df */
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* e0 - ef */
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7  /* f0 - ff */
};

/*
 * rm_class_t *
 * rmc_newclass(...) - Create a new resource management class at priority
 * 'pri' on the interface given by 'ifd'.
 *
 * nsecPerByte  is the data rate of the interface in nanoseconds/byte.
 *              E.g., 800 for a 10Mb/s ethernet.  If the class gets less
 *              than 100% of the bandwidth, this number should be the
 *              'effective' rate for the class.  Let f be the
 *              bandwidth fraction allocated to this class, and let
 *              nsPerByte be the data rate of the output link in
 *              nanoseconds/byte.  Then nsecPerByte is set to
 *              nsPerByte / f.  E.g., 1600 (= 800 / .5)
 *              for a class that gets 50% of an ethernet's bandwidth.
 *
 * action       the routine to call when the class is over limit.
 *
 * maxq         max allowable queue size for class (in packets).
 *
 * parent       parent class pointer.
 *
 * borrow       class to borrow from (should be either 'parent' or null).
 *
 * maxidle      max value allowed for class 'idle' time estimate (this
 *              parameter determines how large an initial burst of packets
 *              can be before overlimit action is invoked.
 *
 * offtime      how long 'delay' action will delay when class goes over
 *              limit (this parameter determines the steady-state burst
 *              size when a class is running over its limit).
 *
 * Maxidle and offtime have to be computed from the following:  If the
 * average packet size is s, the bandwidth fraction allocated to this
 * class is f, we want to allow b packet bursts, and the gain of the
 * averaging filter is g (= 1 - 2^(-RM_FILTER_GAIN)), then:
 *
 *   ptime = s * nsPerByte * (1 - f) / f
 *   maxidle = ptime * (1 - g^b) / g^b
 *   minidle = -ptime * (1 / (f - 1))
 *   offtime = ptime * (1 + 1/(1 - g) * (1 - g^(b - 1)) / g^(b - 1)
 *
 * Operationally, it's convenient to specify maxidle & offtime in units
 * independent of the link bandwidth so the maxidle & offtime passed to
 * this routine are the above values multiplied by 8*f/(1000*nsPerByte).
 * (The constant factor is a scale factor needed to make the parameters
 * integers.  This scaling also means that the 'unscaled' values of
 * maxidle*nsecPerByte/8 and offtime*nsecPerByte/8 will be in microseconds,
 * not nanoseconds.)  Also note that the 'idle' filter computation keeps
 * an estimate scaled upward by 2^RM_FILTER_GAIN so the passed value of
 * maxidle also must be scaled upward by this value.  Thus, the passed
 * values for maxidle and offtime can be computed as follows:
 *
 * maxidle = maxidle * 2^RM_FILTER_GAIN * 8 / (1000 * nsecPerByte)
 * offtime = offtime * 8 / (1000 * nsecPerByte)
 *
 * When USE_HRTIME is employed, then maxidle and offtime become: 
 * 	maxidle = maxilde * (8.0 / nsecPerByte); 
 * 	offtime = offtime * (8.0 / nsecPerByte);
 */

struct rm_class *
rmc_newclass(pri, ifd, nsecPerByte, action, maxq, parent, borrow,
	maxidle, minidle, offtime)
	int		pri;
	struct rm_ifdat	*ifd;
	u_int		nsecPerByte;
	void		(*action)();
	int		maxq;
	struct rm_class	*parent;
	struct rm_class	*borrow;
	u_int		maxidle;
	int		minidle;
	u_int		offtime;
{
	struct rm_class *cl;
	struct rm_class *peer;
	int i, v, tim;

	if (pri >= RM_MAXPRIO)
		return (NULL);

	if (!(cl = (struct rm_class *)kmem_zalloc(sizeof (*cl), KM_SLEEP)))
		return (NULL);
	if (!(cl->q_ = (class_queue_t *)
			kmem_zalloc(sizeof (class_queue_t), KM_SLEEP)))
		return (NULL);

	if (peer = ifd->active_[pri]) {
		/* find the last class at this pri */
		cl->peer_ = peer;
		while (peer->peer_ != ifd->active_[pri])
			peer = peer->peer_;
		peer->peer_ = cl;
	} else {
		ifd->active_[pri] = cl;
		cl->peer_ = cl;
		ifd->hsum_ |= (1 << pri);
	}

	/*
	 * Class initialization.
	 */
	cl->children_ = NULL;
	cl->parent_ = parent;
	cl->borrow_ = borrow;
	cl->leaf_ = 1;
	if (cl->parent_) {
		cl->next_ = parent->children_;
		parent->children_ = cl;
		parent->leaf_ = 0;
	}
	cl->ifdat_ = ifd;
	cl->pri_ = pri;
	cl->allotment_ = RM_NS_PER_SEC / nsecPerByte; /* Bytes per sec */
	cl->depth_ = 0;

	qlimit(cl->q_) = maxq;
	qtype(cl->q_) = Q_DROPTAIL;
	qthresh(cl->q_) = maxq / 20;
	qthresh(cl->q_) = 0;

	/*
	 * Compute the depth of this class and it's ancestors in the class
	 * hierarchy.
	 */
	rmc_depth_compute(cl);

	/*
	 * If CBQ's WRR is enabled, then initailize the class WRR state.
	 */
	if (ifd->wrr_) {
		ifd->num_[pri]++;
		ifd->alloc_[pri] += cl->allotment_;
		rmc_wrr_set_weights(ifd);
	}
	cl->minidle_ = minidle;
#ifdef USE_HRTIME
	if (cl->parent_ == NULL)
		cl->minidle_ = -RM_NS_PER_SEC; 
#endif
	cl->maxidle_ = (maxidle * nsecPerByte) / 8;
	cl->avgidle_ = 0;
	cl->offtime_ = (offtime * nsecPerByte) / 8;
	cl->overlimit = action;

	/*
	 * Pre-compute the packtime times for various size packets.
	 *	XXX - works only for 1500 byte MTU.
	 */
	tim = 0;
	for (i = 0; i < RM_DEFAULT_MTU; i++) {
#ifdef USE_HRTIME
		v = tim;
		cl->len2time_[i] = v;
#else
		v = tim / 1000;
		cl->len2time_[i] =  v == 0? 1 : v;
#endif
		tim += nsecPerByte;
	}
	return (cl);
}

/*
 * static void
 * rmc_wrr_set_weights(struct rm_ifdat *ifdat) - This function computes
 * 	the appropriate run robin weights for the CBQ weighted round robin
 *	algorithm.
 *
 *	Returns: NONE
 */

static void
rmc_wrr_set_weights(struct rm_ifdat *ifd)
{
	int		i;
	struct rm_class	*clh, *cl;

	for (i = 0; i < RM_MAXPRIO; i++) {
		if (ifd->alloc_[i] > 0) {
			/*
			 * This is inverted from that of the simulator to
			 * maintain precision.
			 */
			ifd->M_[i] = ifd->alloc_[i] /
				(ifd->num_[i] * ifd->maxpkt_);

			/*
			 * Compute the weigthed allotment for each class.
			 * This takes the expensive div instruction out
			 * of the main loop for the wrr scheduling path.
			 * These only get recomputed when a class comes or
			 * goes.
			 */
			clh = cl = ifd->active_[i];
			if (cl != NULL) {
				do {
					cl->w_allotment_ = cl->allotment_ /
						ifd->M_[i];
					cl = cl->peer_;
				} while ((cl != NULL) && (cl != clh));
			}
		} else
			ifd->M_[i] = 0;
	}
}

/*
 * static void
 * rmc_depth_compute(struct rm_class *cl) - This function computes the
 * 	appropriate depth of class 'cl' and its ancestors.
 *
 *	Returns:	NONE
 */

static void
rmc_depth_compute(struct rm_class *cl)
{
	rm_class_t *t = cl, *p;

	/*
	 * Recompute the depth for the branch of the tree.
	 */
	while (t != NULL) {
		p = t->parent_;
		if (p && (t->depth_ >= p->depth_)) {
			p->depth_++;
			t = p;
		} else
			t = NULL;
	}
}

/*
 * void
 * rmc_delete_class(struct rm_ifdat *ifdat, struct rm_class *cl) - This
 *	function deletes a class from the link-sharing stucture and frees
 *	all resources associated with the class.
 *
 *	Returns: NONE
 */

void
rmc_delete_class(struct rm_ifdat *ifd, struct rm_class *cl)
{
	struct rm_class		*peer, *head, *previous, *p;
	mbuf_t			*m;

	/*
	 * Free packets in the packet queue
	 */
	if (!qempty(cl->q_)) {
		while ((m = _rmc_getq(cl->q_)) != NULL)
			m_freem(m);

		ifd->na_[cl->pri_]--;
		if (ifd->na_[cl->pri_] == 0)
			ifd->csum_ &= ~(1 << cl->pri_);
	}

	/*
	 * If the class has a parent, then remove the class from the
	 * class from the parent's children chain.
	 */
	if (cl->parent_ != NULL) {
		head = cl->parent_->children_;
		p = previous = head;
		if (head->next_ == NULL) {
			cl->parent_->children_ = NULL;
			cl->parent_->leaf_ = 1;
		} else while (p != NULL) {
			if (p == cl) {
				if (cl == head)
					cl->parent_->children_ = cl->next_;
				else
					previous->next_ = cl->next_;
				cl->next_ = NULL;
				p = NULL;
			} else {
				previous = p;
				p = p->next_;
			}
		}
	}

	/*
	 * Delete class from class priority peer list.
	 */
	if (peer = ifd->active_[cl->pri_]) {
		if (peer != cl) {
			while (peer->peer_ != cl)
				peer = peer->peer_;
			peer->peer_ = cl->peer_;

			if (ifd->active_[cl->pri_] == cl)
				ifd->active_[cl->pri_] = cl->peer_;
		} else {
			ifd->active_[cl->pri_] = NULL;
			ifd->hsum_ &= ~(1 << cl->pri_);
		}
	}

	/*
	 * Recompute the WRR weights.
	 */
	if (ifd->wrr_) {
		ifd->alloc_[cl->pri_] -= cl->allotment_;
		ifd->num_[cl->pri_]--;
		rmc_wrr_set_weights(ifd);
	}

	/*
	 * Free the class structure.
	 */
	kmem_free((void *)cl->q_, sizeof (class_queue_t));
	kmem_free((void *)cl, sizeof (struct rm_class));
}


/*
 * void
 * rmc_init(...) - Initialize the resource management data structures
 *	associated with the output portion of interface 'ifp'.  'ifd' is
 *	where the structures will be built (for backwards compatibility, the
 *	structures aren't kept in the ifnet struct).  'nsecPerByte'
 *	gives the link speed (inverse of bandwidth) in nanoseconds/byte.
 *	'restart' is the driver-specific routine that the generic 'delay
 *	until under limit' action will call to restart output.  `maxq'
 *	is the queue size of the 'link' & 'default' classes.  'maxqueued'
 *	is the maximum number of packets that the resource management
 *	code will allow to be queued 'downstream' (this is typically 1).
 *
 *	Returns:	NONE
 */

void
rmc_init(ifp, ifd, nsecPerByte, restart, maxq, maxqueued, maxidle,
	minidle, offtime, wrr)
#ifdef SOLARIS
	queue_t		*ifp;
#else
	struct ifnet	*ifp;
#endif
	struct rm_ifdat *ifd;
	u_int		nsecPerByte;
	void		(*restart)();
	int		maxq, maxqueued;
	u_int		maxidle;
	int		minidle;
	u_int		offtime;
	int		wrr;
{
	struct rm_class *cl;
	int		i;

	/*
	 * Initialize the CBQ traciing/debug facility.
	 */
	CBQTRACEINIT();	

	bzero((char *)ifd, sizeof (*ifd));
	ifd->ifp = ifp;
	ifd->restart = restart;
	ifd->maxqueued_ = maxqueued;
	ifd->ns_per_byte_ = nsecPerByte;
	ifd->maxpkt_ = 1500; /* XXX - set to default mtusize for e-net. */
	ifd->wrr_ = wrr;

	reset_cutoff(ifd);
	CBQTRACE(rmc_init, 'INIT', ifd->cutoff_);

	/*
	 * Initialize the CBQ's WRR state.
	 */
	for (i = 0; i < RM_MAXPRIO; i++) {
		ifd->alloc_[i] = 0;
		ifd->M_[i] = 0;
		ifd->num_[i] = 0;
		ifd->na_[i] = 0;
		ifd->active_[i] = NULL;
	}

	/*
	 * Initialize current packet state.
	 */
	ifd->qi_ = 0;
	ifd->qo_ = 0;
	for (i = 0; i < RM_MAXQUEUED; i++) {
		ifd->class_[i] = NULL;
		ifd->curlen_[i] = 0;
		ifd->borrowed_[i] = NULL;
		ifd->is_overlimit_[i] = NULL;
	}

	/*
	 * Create the root class of the link-sharing structure.
	 */
	ifd->root_ = rmc_newclass(0, ifd, nsecPerByte, rmc_root_overlimit,
		maxq, 0, 0, maxidle, minidle, offtime);
	if (ifd->root_)
		ifd->root_->depth_ = 0;
}

/*
 * void
 * rmc_queue_packet(struct rm_class *cl, mbuf_t *m) - Add packet given by
 *	mbuf 'm' to queue for resource class 'cl'.  This routine is called
 *	by a driver's if_output routine.  This routine must be called with
 *	output packet completion interrupts locked out (to avoid racing with
 *	rmc_dequeue_next).
 *
 *	Returns:	NONE
 */

void
rmc_queue_packet(struct rm_class *cl, mbuf_t *m)
{
#ifndef USE_HRTIME
	struct timeval	now;
#else
	hrtime_t	now;
#endif
	struct rm_ifdat *ifd = cl->ifdat_;
	int		cpri = cl->pri_;
	int		is_empty = qempty(cl->q_);

	RM_GETTIME(now);
	if (ifd->cutoff_ > 0) {
		if (TV_LT(&cl->undertime_, &now)) {
			ifd->cutoff_ = cl->depth_;
			CBQTRACE(rmc_queue_packet, 'ffoc', cl->depth_);
		} else if ((ifd->cutoff_ > 1) && cl->borrow_) {
			if (TV_LT(&cl->borrow_->undertime_, &now)) {
				ifd->cutoff_ = cl->borrow_->depth_;
				CBQTRACE(rmc_queue_packet, 'ffob',
					cl->borrow_->depth_);
			}
		} 
	}

	_rmc_addq(cl->q_, m);
	if (is_empty) {
		ifd->na_[cpri]++;
		if (ifd->na_[cpri] == 1)
			ifd->csum_ |= (1 << cpri);
	} else  if (qlen(cl->q_) >= qlimit(cl->q_))
		rmc_drop_action(cl);
}

/*
 * void
 * rmc_tl_satisfied(struct rm_ifdat *ifd, struct timeval *now) - Check all
 *	classes to see if there are satified.
 *
 *	Returns:	0, not satisfied.
 *			1, satisfied.
 */

#ifdef USE_HRTIME
void
rmc_tl_satisfied(struct rm_ifdat *ifd, hrtime_t *now)
#else
void
rmc_tl_satisfied(struct rm_ifdat *ifd, struct timeval *now)
#endif
{
	u_int		hsum, i;
	rm_class_t	*p, *bp;

	hsum = ifd->hsum_;
	while (hsum) { 
		i = rmc_mask2pri[hsum];
		if ((bp = ifd->active_[i]) != NULL) {
			p = bp;
			do {
				if (!rmc_satisfied(p, now)) {
					ifd->cutoff_ = p->depth_;
					CBQTRACE(rmc_tl_satisfied, 'ffos',
						p->depth_);
					return;
				} 
				p = p->peer_;
			} while (p->peer_ != bp);
		}
		hsum &= ~(1 << i);
	}
}

/*
 * rmc_satisfied - Return 1 of the class is satisfied.  O, otherwise.
 */

#ifdef USE_HRTIME
static int
rmc_satisfied(struct rm_class *cl, hrtime_t *now)
#else
static int
rmc_satisfied(struct rm_class *cl, struct timeval *now)
#endif
{
	rm_class_t 	*p = cl->children_;

	if (TV_LT(now, &cl->undertime_)) {
		return (1);
	} else if (cl->depth_ == 0) {
		if (!cl->sleeping_ && (qlen(cl->q_) > qthresh(cl->q_)))
			return (0);
		else 
			return (1);
	} else while (p != NULL) {
		if (!rmc_satisfied(p, now))
			return (0);
		p = p->next_;
	}

	return (1);
}

/*
 * Return 1 if class 'cl' is under limit or can borrow from a parent,
 * 0 if overlimit.  As a side-effect, this routine will invoke the
 * class overlimit action if the class if overlimit.
 */
#ifdef USE_HRTIME
int
rmc_under_limit(struct rm_class *cl, hrtime_t *now)
#else
int
rmc_under_limit(struct rm_class *cl, struct timeval *now)
#endif
{
	rm_class_t	*p = cl;
	struct rm_ifdat	*ifd = cl->ifdat_;

	/*
	 * If cl is the root class, then always return that it is
	 * underlimit.  Otherwise, check to see if the class is underlimit.
	 */
	if (cl->parent_ == NULL) {
		return (1);
	} else if (TV_LT(&cl->undertime_, now)) {  
		cl->sleeping_timeout_id_ = 0;
		cl->sleeping_ = 0;
	} else while (TV_LT(now, &cl->undertime_)) {
		++cl->stats_.borrows;
		if (((cl = cl->borrow_) == NULL) ||
				(cl->depth_ > ifd->cutoff_)) {
			if ((cl != NULL) && (cl->depth_ > ifd->cutoff_)) {
				CBQTRACE(rmc_under_limit, 'ffou', ifd->cutoff_);
			}
			p->stats_.overactions++;
			(p->overlimit)(p);
			return (0);
		}
	}
	ifd->borrowed_[ifd->qi_] = cl;
	return (1);
}

/*
 * _rmc_wrr_dequeue_next() - This is scheduler for WRR as opposed to
 *	Packet-by-packet round robin.
 *
 * The heart of the weigthed round-robin scheduler, which decides which
 * class next gets to send a packet.  Highest priority first, then
 * weighted round-robin within priorites.
 *
 * Each able-to-send class gets to send until its byte allocation is
 * exhausted.  Thus, the active pointer is only changed after a class has
 * exhausted its allocation.
 *
 * If the scheduler finds no class that is underlimit or able to borrow,
 * then the first class found that had a nonzero queue and is allowed to
 * borrow gets to send.
 */

static mbuf_t *
_rmc_wrr_dequeue_next(struct rm_ifdat *ifd)
{
	struct rm_class		*clh, *cl = NULL, *first = NULL;
	u_int			csum, cpri, i, deficit = 0;
	int			is_overlimit = 0;
	mbuf_t			*m;
#ifdef USE_HRTIME
	hrtime_t		now;
#else
	struct timeval		now;
#endif

	RM_GETTIME(now);
	csum = ifd->csum_;
	while (csum) {
		/*
		 * Loop through twice for a priority level, if some class
		 * was unable to send a packet the first round because
		 * of the weighted round-robin mechanism.
		 * During the second loop at this level, deficit==2.
		 * (This second loop is not needed if for every class,
		 * "M[cl->pri_])" times "cl->allotment" is greater than
		 * the byte size for the largest packet in the class.)
		 */
wrr_loop:
		cpri = rmc_mask2pri[csum];
		clh = cl = ifd->active_[cpri];
		do {
			if (cl == NULL)
				goto _wrr_next_pri;
			if ((deficit < 2) && (cl->bytes_alloc_ <= 0))
				cl->bytes_alloc_ += cl->w_allotment_;
			if (!qempty(cl->q_)) {
				if (first == NULL && cl->borrow_ != NULL)
					first = cl;
#ifdef USE_HRTIME
				if (rmc_under_limit(cl, &now)) {
#else
				if ((cl->undertime_.tv_sec == 0) ||
						rmc_under_limit(cl, &now)) {
#endif
					if ((cl->bytes_alloc_ > 0 ||
							deficit > 1))
						goto _wrr_out;
					else
						deficit = 1;
				}
			}

			cl->bytes_alloc_ = 0;
			cl = cl->peer_;
		} while (cl && (cl != clh));

		if (deficit == 1) {
			/* Loop on same priority level, with new deficit.  */
			deficit = 2;
			goto wrr_loop;
		} else {
_wrr_next_pri:
			/* New priority level.  */
			deficit = 0;
			csum &= ~(1 << cpri);
		}
	}

	/*
	 * If LINK_EFFICIENCY is turned on, then the first overlimit
	 * class we encounter will send a packet if all the classes
	 * of the link-sharing structure are overlimit.
	 */
	reset_cutoff(ifd);
	if (!ifd->efficient_)
		return (NULL);
	else if (first == NULL)
		return ((mbuf_t *)NULL);
	else {
		is_overlimit++;
		cl = first;
		cpri = cl->pri_;
	}

	/*
	 * Deque the packet and do the book keeping...
	 */
_wrr_out:
	m = _rmc_getq(cl->q_);
	if (qempty(cl->q_)) {
		ifd->na_[cpri]--;
		if (ifd->na_[cpri] == 0)
			ifd->csum_ &= ~(1 << cpri);
	}

	/*
	 * Update class statistics and link data.
	 */
	if (cl->bytes_alloc_ > 0) {
		cl->bytes_alloc_ -= m_len(m);
	}

	if ((cl->bytes_alloc_ <= 0) || is_overlimit)
		ifd->active_[cl->pri_] = cl->peer_;
	else
		ifd->active_[cl->pri_] = cl;

	ifd->class_[ifd->qi_] = cl;
	ifd->curlen_[ifd->qi_] = m_len(m);
	ifd->now_[ifd->qi_] = now;
	ifd->is_overlimit_[ifd->qi_] = is_overlimit;
	ifd->qi_ = (ifd->qi_ + 1) % ifd->maxqueued_;
	ifd->queued_++;

	return (m);
}

/*
 * Dequeue & return next packet from the highest priority class that
 * has a packet to send & has enough allocation to send it.  This
 * routine is called by a driver whenever it needs a new packet to
 * output.
 */
static mbuf_t *
_rmc_prr_dequeue_next(struct rm_ifdat *ifd)
{
	mbuf_t			*m;
	int			is_overlimit = 0, diff, csum, cpri;
	struct rm_class		*clh, *cl, *first = NULL;
#ifdef USE_HRTIME
	hrtime_t		now;
#else
	struct timeval		now;
#endif

	RM_GETTIME(now);
	csum = ifd->csum_;
	while (csum) {
		cpri = rmc_mask2pri[csum];
		clh = cl = ifd->active_[cpri];
		do {
			if (cl == NULL)
				goto _prr_next_pri;
			if (!qempty(cl->q_)) {
				if ((first == NULL) && (cl->borrow_ != NULL))
					first = cl;
#ifdef USE_HRTIME
				if (rmc_under_limit(cl, &now))
#else
				if ((cl->undertime_.tv_sec == 0) ||
						rmc_under_limit(cl, &now))
#endif
					goto _prr_out;
			}
			cl = cl->peer_;
		} while (cl != clh);
_prr_next_pri:
		csum &= ~(1 << cpri);
	}

	/*
	 * If LINK_EFFICIENCY is turned on, then the first overlimit
	 * class we encounter will send a packet if all the classes
	 * of the link-sharing structure are overlimit.
	 */
	reset_cutoff(ifd);
	if (!ifd->efficient_)
		return (NULL);
	else if (first == NULL)
		return ((mbuf_t *)NULL);
	else {
		cl = first;
		cpri = cl->pri_;
		is_overlimit++;
	}

	/*
	 * Deque the packet and do the book keeping...
	 */
_prr_out:
	ifd->active_[cpri] = cl->peer_;
	m = _rmc_getq(cl->q_);
	if (qempty(cl->q_)) {
		ifd->na_[cpri]--;
		if (ifd->na_[cpri] == 0)
			ifd->csum_ &= ~(1 << cpri);
	}

	ifd->class_[ifd->qi_] = cl;
	ifd->curlen_[ifd->qi_] = m_len(m);
	ifd->now_[ifd->qi_] = now;
	ifd->is_overlimit_[ifd->qi_] = is_overlimit;
	ifd->qi_ = (ifd->qi_ + 1) % ifd->maxqueued_;
	ifd->queued_++;

	return (m);
}

/*
 * mbuf_t *
 * rmc_dequeue_next(struct rm_ifdat *ifd, struct timeval *now) - this function
 *	is invoked by the packet driver to get the next packet to be
 *	dequeued and output on the link.  If WRR is enabled, then the
 *	WRR dequeue next routine will determine the next packet to sent.
 *	Otherwise, packet-by-packet round robin is invoked.
 *
 *	Returns:	NULL, if a packet is not available or if all
 *			classes are overlimit.
 *
 *			Otherwise, Pointer to the next packet.
 */

mbuf_t *
rmc_dequeue_next(struct rm_ifdat *ifd)
{
	if (ifd == NULL)
		cmn_err(CE_PANIC, "rmc_dequeue_next, ifd is NULL\n");
	else if (ifd->queued_ >= ifd->maxqueued_)
		return (NULL);
	else if (ifd->wrr_)
		return (_rmc_wrr_dequeue_next(ifd));
	else
		return (_rmc_prr_dequeue_next(ifd));
}

/*
 * Update the utilization estimate for the packet that just completed.
 * The packet's class & the parent(s) of that class all get their
 * estimators updated.  This routine is called by the driver's output-
 * packet-completion interrupt service routine.
 */

void
rmc_update_class_util(struct rm_ifdat *ifd)
{
	int		qlength, idle, avgidle, pktlen, pkttime;
	rm_class_t	*cl, *borrowed;
#ifdef USE_HRTIME
	hrtime_t	now;
#else
	struct timeval	now;
#endif

	/*
	 * Get the most recent completed class.
	 */
	if ((cl = ifd->class_[ifd->qo_]) == NULL)
		return;
	pktlen = ifd->curlen_[ifd->qo_];
	borrowed = ifd->borrowed_[ifd->qo_];
	if (ifd->is_overlimit_[ifd->qo_])
		goto rmc_update_class_util_release;

	/*
	 * Run estimator on class and it's ancesstors.
	 */
	RM_GETTIME(now);
	while (cl != NULL) {
		TV_DELTA(&now, &cl->last_, idle);
		idle -= cl->len2time_[pktlen];
		avgidle = cl->avgidle_;
		avgidle += ((idle - avgidle) >> RM_FILTER_GAIN);
		cl->avgidle_ = avgidle;

		/* Are we overlimit ? */
		if (avgidle <= 0) {
			CBQTRACE(rmc_update_class_util, 'milo',
				cl->stats_.handle);
#ifdef notdef
			if (avgidle < cl->minidle_)
				avgidle = cl->minidle_;
#endif
			TV_ADD_DELTA(&now, cl->len2time_[pktlen],
				&cl->undertime_);
			TV_ADD_DELTA(&cl->undertime_, (1 - RM_POWER) * avgidle,
				&cl->undertime_); 
			++cl->stats_.over;
		} else {
			cl->avgidle_ = (avgidle > cl->maxidle_) ?
				cl->maxidle_ : avgidle;
#ifndef USE_HRTIME
			cl->undertime_.tv_sec = 0;
#endif
		}
		++cl->stats_.npackets;
		cl->stats_.nbytes += pktlen;
		cl->last_ = now;
		cl = cl->parent_;
	} 

	/*
	 * Check to see if cutoff needs to set to a new level.
	 */
	cl = ifd->class_[ifd->qo_];
	if (borrowed && (ifd->cutoff_ >= borrowed->depth_)) {
		if ((qlen(cl->q_) <= 1) || TV_LT(&now, &borrowed->undertime_)) {
			reset_cutoff(ifd);
			CBQTRACE(rmc_update_class_util, 'ffoc', ifd->cutoff_);
		} else if (borrowed != cl) { 
			ifd->cutoff_ = borrowed->depth_;
			CBQTRACE(rmc_update_class_util, 'ffob',
				borrowed->depth_);
		}
	}

	/*
	 * Release class slot
	 */
rmc_update_class_util_release:
	ifd->is_overlimit_[ifd->qo_] = 0;
	ifd->borrowed_[ifd->qo_] = NULL;
	ifd->class_[ifd->qo_] = NULL;
	ifd->qo_ = (ifd->qo_ + 1) % ifd->maxqueued_;
	ifd->queued_--;
}

/*
 * void
 * rmc_drop_action(struct rm_class *cl) - Generic (not protocol-specific)
 *	over-limit action routines.  These get invoked by rmc_under_limit()
 *	if a class with packets to send if over its bandwidth limit & can't
 *	borrow from a parent class.
 *
 *	Returns: NONE
 */

void
rmc_drop_action(struct rm_class *cl)
{
	mbuf_t		*m;
	struct rm_ifdat	*ifd = cl->ifdat_;

	_rmc_dropq(cl->q_);
	if (qempty(cl->q_)) {
		ifd->na_[cl->pri_]--;
		if (ifd->na_[cl->pri_] == 0)
			ifd->csum_ &= ~(1 << cl->pri_);
	}
	++cl->stats_.drops;
}

/*
 * void
 * rmc_delay_action(struct rm_class *cl) - This function is the generic CBQ
 *	delay action routine.  It is invoked via rmc_under_limit when the
 *	packet is discoverd to be overlimit.
 *
 *	Returns: NONE
 */

void
rmc_delay_action(struct rm_class *cl)
{
	int		delay;
#ifdef USE_HRTIME
	hrtime_t	now;
#else
	struct timeval	now;
#endif
	RM_GETTIME(now);
	TV_DELTA(&cl->undertime_, &now, delay);
	delay += cl->offtime_;

	if ((delay > 0) && !cl->sleeping_) {
		CBQTRACE(rmc_delay_action, 'yled', cl->stats_.handle);
		TV_ADD_DELTA(&now, cl->offtime_, &cl->undertime_);
		cl->sleeping_ = 1;
		cl->stats_.delays++;

		/*
	 	 * Since packets are phased randomly with respect to the
	 	 * clock, 1 tick (the next clock tick) can be an arbitrarily
	 	 * short time so we have to wait for at least two ticks.
	 	 * NOTE:  If there's no other traffic, we need the timer as
	 	 * a 'backstop' to restart this class.
	 	 */  
		cl->sleeping_timeout_id_ = timeout(rmc_restart,
			(caddr_t)cl, RM_TIMEOUT); 
	}
}

/*
 * void
 * rmc_restart() - is just a helper routine for rmc_delay_action -- it is
 *	called by the system timer code & is responsible checking if the
 *	class is still sleeping (it might have been restarted as a side
 *	effect of the queue scan on a packet arrival) and, if so, restarting
 *	output for the class.  Inspecting the class state & restarting output
 *	require locking the class structure.  In general the driver is
 *	responsible for locking but this is the only routine that is not
 *	called directly or indirectly from the interface driver so it has
 *	know about system locking conventions.  Under bsd, locking is done
 *	by raising IPL to splimp so that's what's implemented here.  On a
 *	different system this would probably need to be changed.
 *
 *	Returns:	NONE
 */

void
rmc_restart(struct rm_class *cl)
{
	struct rm_ifdat *ifd = cl->ifdat_;

#ifdef SOLARIS
	cbqlock(ifd->ifp);
#endif
	if (cl->sleeping_) {
		if (ifd->queued_ < ifd->maxqueued_)
			(ifd->restart)(ifd->ifp);
		cl->sleeping_ = 0;
		cl->sleeping_timeout_id_ = 0;
	}
#ifdef SOLARIS
	cbqunlock(ifd->ifp);
#endif
}

/*
 * void
 * rmc_root_overlimit(struct rm_class *cl) - This the generic overlimit
 *	handling routine for the root class of the link sharing structure.
 *
 *	Returns: NONE
 */

void
rmc_root_overlimit(struct rm_class *cl)
{
	panic("rmc_root_overlimit");
}

/*
 * Packet Queue handling routines.  Eventually, this is to localize the
 *	effects on the code whether queues are red queues or droptail
 *	queues.
 */

static void
_rmc_addq(class_queue_t *q, mbuf_t *m)
{ 
        mbuf_t  *m0;

	if ((m0 = qtail(q)) != NULL)
		m->m_act = m0->m_act;
	else
		m0 = m;

	m0->m_act = m;
	qtail(q) = m;
	qlen(q)++;
}

static void
_rmc_dropq(class_queue_t *q)
{
        mbuf_t  *m, *m0;

	if (qtail(q) == NULL)
		cmn_err(CE_PANIC, "_rmc_dropq: empty queue");
	else {
		m = qtail(q);
		if (((m0 = m->m_act) != NULL) && (m0 != m))
			m->m_act = m0->m_act;
		else
			qtail(q) = NULL;
	}
	qlen(q)--;
	m_freem(m0);
}

static mbuf_t *
_rmc_getq(class_queue_t *q)
{
        mbuf_t	*m, *m0;

	if (qtail(q) == NULL)
		return (NULL);

	if (((m = qtail(q)) != NULL) && ((m0 = m->m_act) != m)) {
		m->m_act = m0->m_act;
	} else
		qtail(q) = NULL;
	qlen(q)--;
	return (m0);
}
