/*
 * Copyright (c) Sun Microsystems, Inc. 1995-1997. All rights reserved.
 *
 * License is granted to copy, to use, and to make and to use derivative works
 * for research and evaluation purposes, provided that Sun Microsystems is
 * acknowledged in all documentation pertaining to any such copy or
 * derivative work. Sun Microsystems grants no other licenses expressed or
 * implied. The Sun Microsystems  trade name should not be used in any
 * advertising without its written permission.
 *
 * SUN MICROSYSTEMS MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS
 * SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided "as is"
 * without express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this software.
 */

#ifndef _NETINET_RM_CLASS_H
#define	_NETINET_RM_CLASS_H

#pragma ident "@(#)rm_class.h  1.11     97/04/09 SMI"

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Copyright (c) 1991-1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Network Research
 *	Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */


#ifdef _KERNEL

#ifdef RED
#include <netinet/red.h>
#endif

#ifdef SOLARIS
#ifndef _mbuf_
#define	_mbuf_
typedef mblk_t		mbuf_t;
#define	m_act		b_next
#define	m_freem(m0)	freemsg(m0)
#define	m_len(m0)	msgdsize(m0)
#define	microtime	uniqtime
#endif
#endif

/*
 * Macros for dealing with time values.  We assume all times are
 * 'timevals'.  `microtime' is used to get the best available clock
 * resolution.  If `microtime' *doesn't* return a value that's about
 * ten times smaller than the average packet time on the fastest
 * link that will use these routines, a slightly different clock
 * scheme than this one should be used.
 * (Bias due to truncation error in this scheme will overestimate utilization
 * and discriminate against high bandwidth classes.  To remove this bias an
 * integrator needs to be added.  The simplest integrator uses a history of
 * 10 * avg.packet.time / min.tick.time packet completion entries.  This is
 * straight forward to add but we don't want to pay the extra memory
 * traffic to maintain it if it's not necessary (occasionally a vendor
 * accidentally builds a workstation with a decent clock - e.g., Sun & HP).)
 */

#ifndef USE_HRTIME

#define	RM_GETTIME(now) microtime(&now)

#define TV_LT(a, b) (((a)->tv_sec < (b)->tv_sec) ||  \
	(((a)->tv_usec < (b)->tv_usec) && ((a)->tv_sec <= (b)->tv_sec)))

#define	TV_DELTA(a, b, delta) { \
	register int	xxs;	\
							\
	delta = (a)->tv_usec - (b)->tv_usec; \
	if ((xxs = (a)->tv_sec - (b)->tv_sec)) { \
		switch (xxs) { \
		default: \
			if (xxs < 0) \
				cmn_err(CE_CONT, \
					"rm_class: bogus time values"); \
			delta = 0; \
			/* fall through */ \
		case 2: \
			delta += 1000000; \
			/* fall through */ \
		case 1: \
			delta += 1000000; \
			break; \
		} \
	} \
}

#define	TV_ADD_DELTA(a, delta, res) { \
	register int xxus = (a)->tv_usec + (delta); \
	\
	(res)->tv_sec = (a)->tv_sec; \
	while (xxus >= 1000000) { \
		++((res)->tv_sec); \
		xxus -= 1000000; \
	} \
	(res)->tv_usec = xxus; \
}
#else

/*
 * When USE_HRTIME is employed, these macros now use hrtime_t vals
 * for time.  Hrtime_t is a 64-bit representation of wall-clock time on
 * Solaris/Sparc workstations.
 */

#define RM_GETTIME(now)	(now) = gethrtime()

#define TV_LT(a, b)	((*a) < (*b))

#define TV_DELTA(a, b, delta) { \
	(delta) = (*a) - (*b); \
}

#define TV_ADD_DELTA(a, delta, result) { \
	(*result) = (*a) + (delta); \
}

#endif

#define RM_DEFAULT_MTU	1536	/* XXX - Assume ethernet MTU */
#define RM_TIMEOUT	2	/* 1 Clock tick. */

#define	RM_MAXQUEUED	4	/* Max number of packets downstream of CBQ */
#define	RM_MAXPRIO	8	/* Max priority */
#define	RM_MAXQUEUE	64	/* Max queue length */
#define	RM_FILTER_GAIN	5	/* log2 of gain, e.g., 5 => 31/32 */
#define RM_POWER	(2 ^ RM_FILTER_GAIN)
#define	RM_MAXDEPTH	32
#define	RM_NS_PER_SEC	(1000000000)

#ifndef _CLASS_STATS
#define	_CLASS_STATS

typedef struct _class_stats_ {
	u_int		handle;
	u_int		depth;

	int		npackets;	/* packets sent in this class */
	int		nbytes;		/* bytes sent in this class */
	int		over;		/* # times went over limit */
	int		borrows;	/* # times tried to borrow */
	int		drops;		/* # times dropped packets */
	int		overactions;	/* # times invoked overlimit action */
	int		delays;		/* # times invoked delay actions */
} class_stats_t;
#endif

typedef struct rm_ifdat		rm_ifdat_t;
typedef struct rm_class		rm_class_t;
typedef struct _class_queue_	class_queue_t;

/*
 * Packet Queue strcutures and macros to manipulate them.
 */
struct _class_queue_ {
	mbuf_t	*tail_;		/* Tail of packet queue */
	mbuf_t	*head_;		/* Head of packet queue */

	int	qlen_;		/* Queue length (in number of packets) */
	int	qlim_;		/* Queue limit (in number of packets*) */
	int	qthresh_;	/* Queue threshold for formal link sharing */
	int	qtype_;		/* QUeue type */

#ifdef RED
	redq_t	*rqp_;		/* RED Queue state pointer */
#endif
};

/*
 * Packet Queue types: RED or DROPTAIL.
 */
#define Q_DROPTAIL	0x00
#define Q_RED		0x01

#define qtype(q)	(q)->qtype_		/* Get queue type */
#define qthresh(q)	(q)->qthresh_		/* Threshold for satifaction */
#define qlimit(q)	(q)->qlim_		/* Max packets to be queued */
#define	qlen(q)		(q)->qlen_		/* Current queue length. */
#define	qtail(q)	(q)->tail_		/* Tail of the queue */
#define qhead(q)	(q)->head_		/* Head of the queue */

#define qempty(q)	((q)->qlen_ == 0)	/* Is the queue empty?? */
#define q_is_red(q)	((q)->qtype_ == Q_RED)	/* Is the queue a red queue */

/*
 * CBQ Class state structure
 */
struct rm_class {
	class_queue_t	*q_;		/* Queue of packets */
	rm_ifdat_t	*ifdat_;			
	int		pri_;		/* Class priority. */
	int		depth_;		/* Class depth */
	u_int		maxrate_;	/* Bytes per second for this class. */
	u_int		allotment_;	/* Fraction of link bandwidth. */
	u_int		w_allotment_;	/* Weighted allotment for WRR */
	int		bytes_alloc_;	/* Allocation for round of WRR */

	int		avgidle_;
	int		maxidle_;
	int		minidle_;
	int		offtime_;
	int		sleeping_;	/* != 0 if delaying */
	int		sleeping_timeout_id_;	

#ifdef USE_HRTIME
	hrtime_t	last_;
	hrtime_t	undertime_;
#else
	struct timeval	last_;		/* time last packet sent */
	struct timeval	undertime_;	/* time can next send */
#endif
	int		leaf_;		/* Note whether leaf class or not.*/

	rm_class_t	*children_;	/* Children of this class */
	rm_class_t	*next_;		/* Next pointer, used if child */

	rm_class_t	*peer_;		/* Peer class */
	rm_class_t	*borrow_;	/* Borrow class */
	rm_class_t	*parent_;	/* Parent class */

	void		(*overlimit)();	/* Over limit action. */
	void		(*drop)();	/* Class drop action. */

	class_stats_t	stats_;		/* Class Statistics */

	int		len2time_[RM_DEFAULT_MTU];	/* Packet times */
};

/*
 * CBQ Interface state
 */
struct rm_ifdat {
	int		queued_;	/* # pkts queued downstream */
	int		efficient_;	/* Link Efficency bit */
	int		wrr_;		/* Enable Weighted Round-Robin */
	u_long		ns_per_byte_;	/* Link byte speed. */
	int		maxqueued_;	/* Max packets to queue */
	int		maxpkt_;	/* Max packet size. */
	int		qi_;		/* In/out pointers for downstream */
	int		qo_;		/* packets */

	/*
	 * Active class state and WRR state.
	 */
	rm_class_t	*active_[RM_MAXPRIO];	/* Active cl's in each pri */
	int		na_[RM_MAXPRIO];	/* # of active cl's in a pri */
	rm_class_t	*classes_[RM_MAXPRIO];	/* All classes in pri */
	int		num_[RM_MAXPRIO];	/* # of cl's per pri */
	int		alloc_[RM_MAXPRIO];	/* Byte Allocation */
	u_long		M_[RM_MAXPRIO];		/* WRR weights. */

	/*
	 * Network Interface/Solaris Queue state pointer.
	 */
#ifdef SOLARIS
	queue_t		*ifp;
#else
	struct ifnet	*ifp;
#endif
	rm_class_t	*default_;	/* Default Pkt class, BE */
	rm_class_t	*root_;		/* Root Link class. */
	rm_class_t	*ctl_;		/* Control Traffic class. */
	void		(*restart)();	/* Restart routtine. */

	/*
	 * Current packet downstream packet state and dynamic state.
	 */
	rm_class_t	*borrowed_[RM_MAXQUEUED]; /* Class borrowed last */
	rm_class_t	*class_[RM_MAXQUEUED];	/* class sending */
	int		curlen_[RM_MAXQUEUED];	/* Current pktlen */
#ifndef USE_HRTIME
	struct timeval	now_[RM_MAXQUEUED];	/* Current packet time. */
#else
	hrtime_t	now_[RM_MAXQUEUED];	/* Current packet time. */
#endif
	int		is_overlimit_[RM_MAXQUEUED];/* Current packet time. */

	int		cutoff_;	/* Cut-off depth for borrowing */
	int		csum_;		/* Class priorites with active cl's */
	int		hsum_;		/* Class priorites that exist in the */
					/* class hierararchy. */
};

#ifdef SOLARIS

extern void		rmc_init(queue_t *q, struct rm_ifdat *ifdat,
				u_int nsecPerByte, void (*restart)(),
				int maxq, int maxqueued, u_int maxidle,
				int minidle, u_int offtime, int wrr);
#ifndef USE_HRTIME
extern int		rmc_is_under_limit(struct rm_class *, struct timeval *);
extern int		rmc_under_limit(struct rm_class *, struct timeval *);
extern void		rmc_tl_satisfied(struct rm_ifdat *, struct timeval *);
#else
extern int		rmc_is_under_limit(struct rm_class *, hrtime_t *);
extern int		rmc_under_limit(struct rm_class *, hrtime_t *);
extern void		rmc_tl_satisfied(struct rm_ifdat *, hrtime_t *);
#endif

extern void		rmc_drop_action(struct rm_class *);
extern void		rmc_delay_action(struct rm_class *);
extern void		rmc_queue_packet(struct rm_class *, mbuf_t *);
extern mbuf_t		*rmc_dequeue_next(struct rm_ifdat *);
extern void		rmc_update_class_util(struct rm_ifdat *);
extern rm_class_t	*rmc_newclass(int pri, struct rm_ifdat *ifdat,
				u_int nsecPerByte, void (*action)(),
				int maxq, struct rm_class *parent,
				struct rm_class *borrow, u_int maxidle,
				int minidle, u_int offtime);
extern void		rmc_delete_class(struct rm_ifdat *, struct rm_class *);
extern void		rmc_restart(struct rm_class *);
extern void		rmc_root_overlimit(struct rm_class *);

#endif /* SOLARIS */
#endif /* _KERNEL */

#ifdef __cplusplus
}
#endif

#endif /* !NETINET_RM_CLASS_H */
